# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


import logging
from typing import Optional, Tuple, Union

import pandas as pd

from ..base import (
    Artifact,
    EventDataFrame,
    InSamplePredictions,
    Pipeline,
    PipelineCard,
    TrainedPipelineCard,
)
from ..events import _create_events
from ..splitters import Fold, SlidingWindowSplitter, Splitter
from ..utils.dataframe import concat_on_index_override_duplicate_rows
from ..utils.list import unpack_list_of_tuples, wrap_in_list
from .backend import get_backend
from .checks import check_types
from .common import _sequential_train_on_window, _train_on_window
from .types import Backend, BackendType, TrainMethod
from .utils import _extract_trained_pipelines
from .wrap import wrap_transformation_if_needed

logger = logging.getLogger("fold:loop")


def train(
    pipelinecard: Union[PipelineCard, Pipeline],
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    events: Optional[EventDataFrame] = None,
    train_method: Union[TrainMethod, str] = TrainMethod.parallel,
    backend: Union[BackendType, Backend, str] = BackendType.no,
    silent: bool = False,
    return_artifacts: bool = False,
    return_insample: bool = False,
    for_deployment: bool = False,
) -> Union[
    Tuple[TrainedPipelineCard],
    Tuple[TrainedPipelineCard, Artifact],
    Tuple[TrainedPipelineCard, Artifact, InSamplePredictions],
]:
    """
    Trains a pipeline on a given dataset, for all folds returned by the Splitter.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to be fitted.
    X: Optional[pd.DataFrame]
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    train_method : TrainMethod, str = TrainMethod.parallel
        The training methodology, by default `parallel`.
    backend: str, BackendType = BackendType.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    events: EventDataFrame, optional = None
        Events that should be passed into the pipeline, by default None.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    return_artifacts: bool = False
        Whether to return the artifacts of the training process, by default False.
    return_insample: bool = False
        Whether to return the in-sample predictions of the training process, by default False.
    for_deployment: bool = False
        Whether the pipeline is being trained for deployment, meaning it'll only have the last fold, by default False.

    Returns
    -------
    TrainedPipelineCard
        The fitted pipelines, for all folds.
    """
    pipelinecard = (
        pipelinecard
        if isinstance(pipelinecard, PipelineCard)
        else PipelineCard(preprocessing=None, pipeline=pipelinecard)
    )
    X, y = check_types(X, y)
    if events is None:
        events = _create_events(y, pipelinecard)
    if events is not None and events.shape[0] != X.shape[0]:
        logger.warning("The number of events does not match the number of samples.")
        events = events.reindex(X.index)
    artifact = Artifact.from_events(X.index, events)
    train_method = TrainMethod.from_str(train_method)
    backend = get_backend(backend)

    if pipelinecard.preprocessing is not None:
        preprocessing_pipeline = wrap_in_list(pipelinecard.preprocessing)
        preprocessing_pipeline = wrap_transformation_if_needed(preprocessing_pipeline)
        (
            _,
            trained_preprocessing_pipeline,
            preprocessed_X,
            preprocessed_artifact,
        ) = _train_on_window(
            X=X,
            y=y,
            artifact=artifact,
            pipeline=preprocessing_pipeline,
            split=Fold(0, 0, 0, len(X), 0, 0, 0, len(X)),
            never_update=True,
            backend=backend,
            project_name=f"{pipelinecard.name}-Preprocessing"
            if pipelinecard.name is not None
            else "Preprocessing",
            show_progress=True,
        )
        assert preprocessed_X.shape[0] == X.shape[0]
        assert preprocessed_artifact.shape[0] == artifact.shape[0]
        X = preprocessed_X
        artifact = preprocessed_artifact

    if isinstance(splitter, SlidingWindowSplitter):
        assert train_method != TrainMethod.sequential, (
            "SlidingWindowSplitter is conceptually incompatible with"
            " TrainMethod.sequential"
        )

    pipeline = wrap_in_list(pipelinecard.pipeline)
    pipeline = wrap_transformation_if_needed(pipeline)

    splits = splitter.splits(index=X.index)
    if len(splits) == 0:
        raise ValueError("No splits were generated by the Splitter.")

    if for_deployment:
        (
            processed_idx,
            processed_pipelines,
            processed_predictions,
            processed_artifacts,
        ) = unpack_list_of_tuples(
            backend.train_pipeline(
                _train_on_window,
                pipeline,
                X,
                y,
                artifact,
                splits[-1:],
                True,
                backend,
                f"{pipelinecard.name}-Pipeline" or "Pipeline",
                silent,
            )
        )

    elif train_method == TrainMethod.parallel_with_search and len(splits) > 1:
        (
            first_batch_index,
            first_batch_transformations,
            first_batch_predictions,
            first_batch_artifacts,
        ) = _train_on_window(
            X=X,
            y=y,
            artifact=artifact,
            pipeline=pipeline,
            split=splits[0],
            never_update=True,
            backend=backend,
            project_name=f"{pipelinecard.name}-Pipeline" or "Pipeline",
        )

        (
            rest_idx,
            rest_transformations,
            rest_predictions,
            rest_artifacts,
        ) = unpack_list_of_tuples(
            backend.train_pipeline(
                _train_on_window,
                first_batch_transformations,
                X,
                y,
                artifact,
                splits[1:],
                False,
                backend,
                f"{pipelinecard.name}-Pipeline" or "Pipeline",
                silent,
            )
        )
        processed_idx = [first_batch_index] + list(rest_idx)
        processed_pipelines = [first_batch_transformations] + list(rest_transformations)
        processed_predictions = [first_batch_predictions] + list(rest_predictions)
        processed_artifacts = [first_batch_artifacts] + list(rest_artifacts)

    elif train_method == TrainMethod.parallel:
        (
            processed_idx,
            processed_pipelines,
            processed_predictions,
            processed_artifacts,
        ) = unpack_list_of_tuples(
            backend.train_pipeline(
                _train_on_window,
                pipeline,
                X,
                y,
                artifact,
                splits,
                True,
                backend,
                f"{pipelinecard.name}-Pipeline" or "Pipeline",
                silent,
            )
        )

    else:
        (
            processed_idx,
            processed_pipelines,
            processed_predictions,
            processed_artifacts,
        ) = _sequential_train_on_window(
            pipeline,
            X,
            y,
            splits,
            artifact,
            backend,
            project_name=f"{pipelinecard.name}-Pipeline" or "Pipeline",
        )

    trained_pipelines = TrainedPipelineCard(
        name=pipelinecard.name or "",
        preprocessing=[pd.Series(p, index=[0]) for p in trained_preprocessing_pipeline]
        if pipelinecard.preprocessing
        else None,
        pipeline=_extract_trained_pipelines(processed_idx, processed_pipelines),
        event_labeler=pipelinecard.event_labeler,
        event_filter=pipelinecard.event_filter,
    )
    if return_artifacts is True:
        processed_artifacts = concat_on_index_override_duplicate_rows(
            processed_artifacts
        )
        processed_predictions = concat_on_index_override_duplicate_rows(
            processed_predictions
        )
        assert processed_artifacts.index.equals(processed_predictions.index)
        if return_insample:
            return (
                trained_pipelines,
                processed_artifacts,
                processed_predictions,
            )
        else:
            return trained_pipelines, processed_artifacts
    else:
        if return_insample:
            return trained_pipelines, concat_on_index_override_duplicate_rows(
                processed_predictions
            )
        else:
            return trained_pipelines



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from abc import abstractmethod
from typing import Union

import numpy as np
import pandas as pd

from ..base import Transformation
from ..utils.checks import (
    get_classes_from_probabilies_column_names,
    get_probabilities_column_names,
    has_probabilities,
)


class Model(Transformation):
    @abstractmethod
    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for exclusively out-of-sample data, the model has never seen before.
        """
        raise NotImplementedError

    @abstractmethod
    def predict_in_sample(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for in-sample, already seen data.
        """
        raise NotImplementedError

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            return postpostprocess_output(self.predict_in_sample(X), self.name)
        else:
            return postpostprocess_output(self.predict(X), self.name)


class TimeSeriesModel(Transformation):
    """
    Convenience class, for models that have access to past `y` values.
    """

    @abstractmethod
    def predict(
        self,
        X: pd.DataFrame,
        past_y: pd.Series,
    ) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for exclusively out-of-sample data, the model has never seen before.
        """
        raise NotImplementedError

    @abstractmethod
    def predict_in_sample(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for in-sample, already seen data.
        """
        raise NotImplementedError

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            return postpostprocess_output(
                self.predict_in_sample(X, self._state.memory_y.shift(1)), self.name
            )
        else:
            return postpostprocess_output(
                self.predict(
                    X[-(self.properties.memory_size + 1) : None],
                    pd.Series(
                        np.concatenate(
                            [
                                np.ones((1,)) * np.nan,
                                self._state.memory_y[
                                    -self.properties.memory_size : None
                                ],
                            ]
                        ),
                        index=X.index,
                    ),
                ),
                self.name,
            )


def postpostprocess_output(
    dataframe_or_series: Union[pd.DataFrame, pd.Series], name: str
) -> pd.DataFrame:
    if isinstance(dataframe_or_series, pd.Series):
        return dataframe_or_series.rename(f"predictions_{name}").to_frame()
    elif isinstance(dataframe_or_series, pd.DataFrame):
        if len(dataframe_or_series.columns) == 1:
            dataframe_or_series.columns = [f"predictions_{name}"]
            return dataframe_or_series
        elif has_probabilities(dataframe_or_series):
            # Sort probabilities columns if necessary
            prob_columns = get_probabilities_column_names(dataframe_or_series)
            prob_columns_sorted = sorted(
                get_classes_from_probabilies_column_names(prob_columns)
            )
            if prob_columns == prob_columns_sorted:
                return dataframe_or_series
            else:
                return dataframe_or_series.sort_index(axis="columns")
        else:
            return dataframe_or_series
    else:
        raise ValueError(
            f"Expected dataframe or series, got {type(dataframe_or_series)} instead."
        )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from .backtesting import *
from .encase import *
from .inference import *
from .training import *
from .types import *
from .updating import *



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

from typing import Tuple

from sklearn.preprocessing import MinMaxScaler as SKLearnMinMaxScaler
from sklearn.preprocessing import StandardScaler as SKLearnStandardScaler

from fold.transformations.sklearn import WrapInvertibleSKLearnTransformation


class StandardScaler(WrapInvertibleSKLearnTransformation):
    """
    Standardize features by removing the mean and scaling to unit variance.

    A wrapper around SKLearn's StandardScaler.
    Capable of further updates after the initial fit.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import StandardScaler
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = StandardScaler()
    >>> X["sine"].head()
    2021-12-31 07:20:00    0.0000
    2021-12-31 07:21:00    0.0126
    2021-12-31 07:22:00    0.0251
    2021-12-31 07:23:00    0.0377
    2021-12-31 07:24:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds["sine"].head()
    2021-12-31 15:40:00   -0.000000
    2021-12-31 15:41:00    0.017819
    2021-12-31 15:42:00    0.035497
    2021-12-31 15:43:00    0.053316
    2021-12-31 15:44:00    0.070994
    Freq: T, Name: sine, dtype: float64

    ```

    References
    ----------

    [SKLearn's StandardScaler documentation](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html)
    """

    name = "StandardScaler"

    def __init__(self):
        super().__init__(SKLearnStandardScaler, init_args=dict())


class MinMaxScaler(WrapInvertibleSKLearnTransformation):
    """
    Transform features by scaling each feature to a given range.

    A wrapper around SKLearn's StandardScaler.
    Capable of further updates after the initial fit.

    Parameters
    ----------
    feature_range : tuple (min, max), default=(0, 1)
        Desired range of transformed data.

    clip : bool, default=False
        Set to True to clip transformed values of held-out data to
        provided `feature range`.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import MinMaxScaler
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = MinMaxScaler()
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> X["sine"].loc[preds.index].head()
    2021-12-31 15:40:00   -0.0000
    2021-12-31 15:41:00    0.0126
    2021-12-31 15:42:00    0.0251
    2021-12-31 15:43:00    0.0377
    2021-12-31 15:44:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds["sine"].head()
    2021-12-31 15:40:00    0.50000
    2021-12-31 15:41:00    0.50630
    2021-12-31 15:42:00    0.51255
    2021-12-31 15:43:00    0.51885
    2021-12-31 15:44:00    0.52510
    Freq: T, Name: sine, dtype: float64

    ```

    References
    ----------
    [SKLearn's MinMaxScaler documentation](https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html)
    """

    name = "MinMaxScaler"

    def __init__(self, feature_range: Tuple[int, int] = (0, 1), clip=False):
        super().__init__(
            SKLearnMinMaxScaler, init_args=dict(feature_range=feature_range, clip=clip)
        )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Optional, Tuple

import pandas as pd


def check_types(
    X: Optional[pd.DataFrame], y: pd.Series
) -> Tuple[pd.DataFrame, pd.Series]:
    if X is None:
        X = pd.DataFrame(
            pd.arrays.SparseArray(0), index=y.index, columns=["X_not_available"]
        )
    else:
        assert isinstance(X, pd.DataFrame), "X must be a pandas DataFrame."
    assert isinstance(y, pd.Series), "y must be a pandas Series."
    return X, y


def check_types_multi_series(data: pd.DataFrame):
    assert isinstance(data, pd.DataFrame), "data must be a pandas DataFrame."
    assert "y" in data.columns, "data must have a column `y`."
    assert "ds" in data.columns, "data must have a column `ds`."
    assert "unique_id" in data.columns, "data must have a column `unique_id`."



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

from __future__ import annotations

from inspect import getfullargspec
from typing import Callable, Optional, Type

import numpy as np
import pandas as pd

from ..base import (
    Artifact,
    FeatureSelector,
    InvertibleTransformation,
    Transformation,
    Tunable,
    fit_noop,
)


class WrapSKLearnTransformation(Transformation, Tunable):
    """
    Wraps an SKLearn Transformation.
    There's no need to use it directly, `fold` automatically wraps all sklearn transformations into this class.
    """

    def __init__(
        self,
        transformation_class: Type,
        init_args: dict,
        output_dtype: Optional[type] = None,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.transformation_class = transformation_class
        self.init_args = init_args
        self.transformation = transformation_class(**init_args)
        if hasattr(self.transformation, "set_output"):
            self.transformation = self.transformation.set_output(transform="pandas")
        self.params_to_try = params_to_try
        self.output_dtype = output_dtype
        self.name = name or self.transformation.__class__.__name__
        self.properties = Transformation.Properties(requires_X=True)

    @classmethod
    def from_model(
        cls,
        model,
        output_dtype: Optional[type] = None,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> WrapSKLearnTransformation:
        return cls(
            transformation_class=model.__class__,
            init_args=model.get_params(deep=False),
            output_dtype=output_dtype,
            name=name,
            params_to_try=params_to_try,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        fit_func = (
            self.transformation.partial_fit
            if hasattr(self.transformation, "partial_fit")
            else self.transformation.fit
        )

        argspec = getfullargspec(fit_func)
        if len(argspec.args) == 2 or len(argspec.args) == 1:
            fit_func(X)
        elif len(argspec.args) == 3:
            fit_func(X, y)
        elif len(argspec.args) == 4:
            fit_func(X, y, sample_weights)
        else:
            raise ValueError(
                f"Unexpected number of arguments in {self.transformation.__class__.__name__}.fit"
            )

    def update(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        if not hasattr(self.transformation, "partial_fit"):
            return
        argspec = getfullargspec(self.transformation.partial_fit)

        if len(argspec.args) == 2:
            self.transformation.partial_fit(X)
        elif len(argspec.args) == 3:
            self.transformation.partial_fit(X, y)
        elif len(argspec.args) == 4:
            self.transformation.partial_fit(X, y, sample_weights)
        else:
            raise ValueError(
                f"Unexpected number of arguments in {self.transformation.__class__.__name__}.partial_fit"
            )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        def convert_dtype_if_needed(df: pd.DataFrame) -> pd.DataFrame:
            return df.astype(self.output_dtype) if self.output_dtype is not None else df

        result = convert_dtype_if_needed(self.transformation.transform(X))
        if hasattr(self.transformation, "set_output"):
            return result
        else:
            if result.shape[1] != len(X.columns):
                columns = [f"{self.name}_{i}" for i in range(result.shape[1])]
            else:
                columns = X.columns
            return pd.DataFrame(result, index=X.index, columns=columns)

    def get_params(self) -> dict:
        return self.transformation.get_params(deep=False)

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        return WrapSKLearnTransformation(
            transformation_class=self.transformation.__class__,
            output_dtype=self.output_dtype,
            init_args=parameters,
            name=self.name,
        )


class WrapInvertibleSKLearnTransformation(
    WrapSKLearnTransformation, InvertibleTransformation
):
    def inverse_transform(self, X: pd.Series, in_sample: bool) -> pd.Series:
        return pd.Series(self.transformation.inverse_transform(X), index=X.index)


class WrapSKLearnFeatureSelector(FeatureSelector, Tunable):
    """
    Wraps an SKLearn Feature Selector class, stores the selected columns in `selected_features` property.
    There's no need to use it directly, `fold` automatically wraps all sklearn feature selectors into this class.
    """

    selected_features: Optional[str] = None

    def __init__(
        self,
        transformation_class: Type,
        init_args: dict,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.transformation = transformation_class(**init_args)
        self.transformation_class = transformation_class
        self.init_args = init_args
        if hasattr(self.transformation, "set_output"):
            self.transformation = self.transformation.set_output(transform="pandas")
        self.params_to_try = params_to_try
        self.name = name or self.transformation.__class__.__name__
        self.properties = Transformation.Properties(requires_X=True)

    @classmethod
    def from_model(
        cls, model, name: Optional[str] = None, params_to_try: Optional[dict] = None
    ) -> WrapSKLearnFeatureSelector:
        return cls(
            transformation_class=model.__class__,
            init_args=model.get_params(deep=False),
            name=name,
            params_to_try=params_to_try,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        self.transformation.fit(X, y)
        if hasattr(self.transformation, "get_feature_names_out"):
            self.selected_features = self.transformation.get_feature_names_out()
        else:
            self.selected_features = X.columns[
                self.transformation.get_support()
            ].to_list()
        return pd.DataFrame(
            {f"selected_features_{self.name}": [self.selected_features]},
            index=X.index[-1:],
        ).reindex(X.index)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X[self.selected_features]

    def get_params(self) -> dict:
        return self.transformation.get_params(deep=False)

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        return WrapSKLearnFeatureSelector(
            transformation_class=self.transformation.__class__,
            init_args=parameters,
            name=self.name,
        )

    update = fit_noop


class RemoveLowVarianceFeatures(FeatureSelector):
    def __init__(self, threshold: float = 1e-5, name: Optional[str] = None) -> None:
        self.threshold = threshold
        self.properties = Transformation.Properties(requires_X=True)
        self.name = name or "RemoveLowVarianceFeatures"

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        self.variances_ = np.nanvar(X, axis=0)
        if self.threshold == 0:
            peak_to_peaks = np.ptp(X, axis=0)
            compare_arr = np.array([self.variances_, peak_to_peaks])
            self.variances_ = np.nanmin(compare_arr, axis=0)

        if np.all(~np.isfinite(self.variances_) | (self.variances_ <= self.threshold)):
            msg = "No feature in X meets the variance threshold {0:.5f}"
            if X.shape[0] == 1:
                msg += " (X contains only one sample)"
            raise ValueError(msg.format(self.threshold))
        mask = self.variances_ > self.threshold
        self.selected_features = X.columns[mask].to_list()

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X[self.selected_features]

    update = fit_noop



from __future__ import annotations

from typing import Any, Optional, Type, Union

import pandas as pd

from fold.models.base import Model
from fold.utils.checks import is_X_available


class WrapStatsModels(Model):
    def __init__(
        self,
        model_class: Type,
        init_args: dict,
        use_exogenous: Optional[bool] = None,
        online_mode: bool = False,
        instance: Optional[Any] = None,
        name: Optional[str] = None,
    ) -> None:
        self.init_args = init_args
        init_args = {} if init_args is None else init_args
        self.model_class = model_class
        self.use_exogenous = use_exogenous
        self.properties = Model.Properties(
            requires_X=False,
            model_type=Model.Properties.ModelType.regressor,
            mode=(
                Model.Properties.Mode.online
                if online_mode
                else Model.Properties.Mode.minibatch
            ),
        )
        self.name = name or self.model_class.__class__.__name__
        self.instance = instance

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            self.model = (
                self.model_class(y, X, **self.init_args)
                if self.instance is None
                else self.instance
            )
            self.model = self.model.fit()
        else:
            self.model = (
                self.model_class(y, **self.init_args)
                if self.instance is None
                else self.instance
            )
            self.model = self.model.fit()

    def update(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        if not hasattr(self.model, "append"):
            return
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            self.model = self.model.append(endog=y, exog=X, refit=True)
        else:
            self.model = self.model.append(endog=y, refit=True)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            return pd.Series(
                self.model.predict(start=X.index[0], end=X.index[-1], exog=X)
            )
        else:
            return pd.Series(self.model.predict(start=X.index[0], end=X.index[-1]))

    def predict_in_sample(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        return self.model.predict(start=X.index[0], end=X.index[-1])



from fold.composites import ModelResiduals, TransformTarget
from fold.transformations.dev import Identity

from .ar import AR


class ARCH(TransformTarget):
    def __init__(self, p: int):
        super().__init__(
            wrapped_pipeline=AR(p),
            y_pipeline=lambda x: x**2,
            invert_wrapped_output=False,
        )


class GARCH(TransformTarget):
    """
    GARCH(p, q) model.
    Accepts a stationary time series (eg. returns), predicts the variance of the time series.

    CAUTION:
    The model assumes that the mean of the time series is 0.0.

    Parameters
    ----------
    p : int
        The number of lags to include in the model.
    q : int
        The number of lags to include in the residual model.
    """

    def __init__(self, p: int, q: int, ma_model_online: bool = False):
        assert p >= 0, "p must be above 0"
        assert q >= 0, "q must be above 0"

        if q > 0:
            ma_model = AR(q)
            ma_model.properties._internal_supports_minibatch_backtesting = (
                not ma_model_online
            )
        else:
            ma_model = Identity()

        garch = ModelResiduals(
            primary=AR(p),
            meta=ma_model,
        )
        super().__init__(
            wrapped_pipeline=garch,
            y_pipeline=lambda x: x**2,
            invert_wrapped_output=False,
        )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.
from __future__ import annotations

from typing import Callable, List, Optional, Tuple, Union

import pandas as pd

from fold.base.classes import TrainedPipelineCard

from ..base import (
    Artifact,
    Composite,
    EventDataFrame,
    EventFilter,
    Labeler,
    LabelingStrategy,
    Pipeline,
    PipelineCard,
    Pipelines,
    T,
    WeightingStrategy,
)
from ..utils.list import wrap_in_double_list_if_needed
from .filters import EveryNth, FilterZero, NoFilter
from .labeling import *  # noqa
from .weights import *  # noqa


def _create_events(
    y: pd.Series, pipeline_card: Union[PipelineCard, TrainedPipelineCard]
) -> Optional[EventDataFrame]:
    if pipeline_card.event_labeler is None:
        return None
    start_times = (
        pipeline_card.event_filter.get_event_start_times(y)
        if pipeline_card.event_filter is not None
        else y.index
    )
    return pipeline_card.event_labeler.label_events(start_times, y).reindex(y.index)



from __future__ import annotations

from typing import Any, Dict, Optional, Type, Union

import numpy as np
import pandas as pd

from fold.models.base import Model


class WrapProphet(Model):
    def __init__(
        self,
        model_class: Type,
        init_args: Optional[Dict],
        online_mode: bool = False,
        instance: Optional[Any] = None,
    ) -> None:
        self.init_args = init_args
        init_args = {} if init_args is None else init_args
        self.model = model_class(**init_args) if instance is None else instance
        self.model_class = model_class
        self.properties = Model.Properties(
            requires_X=False,
            model_type=Model.Properties.ModelType.regressor,
            mode=(
                Model.Properties.Mode.online
                if online_mode
                else Model.Properties.Mode.minibatch
            ),
        )
        self.name = self.model_class.__class__.__name__

    @classmethod
    def from_model(
        cls,
        model,
        online_mode: bool = False,
    ) -> WrapProphet:
        return cls(
            model_class=model.__class__,
            init_args={},
            instance=model,
            online_mode=online_mode,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        data = pd.DataFrame({"ds": X.index, "y": y.values})
        self.model.fit(data)

    def update(
        self,
        X: pd.DataFrame,
        y: Optional[pd.Series],
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if self.init_args is None:
            raise ValueError(
                "Cannot update model if init_args is None, probably .from_model"
                " constructor was used."
            )
        data = pd.DataFrame({"ds": X.index, "y": y.values})
        old_model = self.model
        self.model = self.model_class(**self.init_args)
        self.model.fit(data, init=warm_start_params(old_model))

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        data = pd.DataFrame({"ds": X.index})
        predictions = self.model.predict(data)
        return pd.Series(predictions["yhat"].values, index=X.index)

    predict_in_sample = predict


def warm_start_params(m):
    """
    # from https://facebook.github.io/prophet/docs/additional_topics.html
    Retrieve parameters from a trained model in the format used to initialize a new Stan model.
    Note that the new Stan model must have these same settings:
        n_changepoints, seasonality features, mcmc sampling
    for the retrieved parameters to be valid for the new model.

    Parameters
    ----------
    m: A trained model of the Prophet class.

    Returns
    -------
    A Dictionary containing retrieved parameters of m.
    """
    res = {}
    for pname in ["k", "m", "sigma_obs"]:
        if m.mcmc_samples == 0:
            res[pname] = m.params[pname][0][0]
        else:
            res[pname] = np.mean(m.params[pname])
    for pname in ["delta", "beta"]:
        if m.mcmc_samples == 0:
            res[pname] = m.params[pname][0]
        else:
            res[pname] = np.mean(m.params[pname], axis=0)
    return res



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from random import choices
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from fold.base.classes import Artifact

from ..base import Transformation, fit_noop
from .base import Model


class RandomClassifier(Model):
    """
    A model that predicts random classes and probabilities.

    Parameters
    ----------
    all_classes : List[int]
        All possible classes.
    probability_mean : float
        The mean of the normal distribution used to generate the probabilities.


    Examples
    --------
    ```pycon
    >>> import numpy as np
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.models import RandomClassifier
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> np.random.seed(42)
    >>> pipeline = RandomClassifier([0,1], [0.5, 0.5])
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)

    ```
    """

    name = "RandomClassifier"

    def __init__(
        self, all_classes: List[int], probability_mean: Optional[List[float]] = None
    ) -> None:
        self.all_classes = all_classes
        if probability_mean is not None:
            assert len(probability_mean) == len(all_classes)
        self.probability_mean = (
            [1 / len(all_classes) for _ in range(len(all_classes))]
            if probability_mean is None
            else probability_mean
        )
        self.properties = Transformation.Properties(requires_X=False)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        predictions = pd.Series(
            choices(population=self.all_classes, k=len(X)),
            index=X.index,
            name="predictions_RandomClassifier",
        )
        probabilities = pd.concat(
            [
                pd.Series(
                    np.random.normal(prob_mean, 0.1, len(X)).clip(0, 1),
                    index=X.index,
                    name=f"probabilities_RandomClassifier_{associated_class}",
                )
                for associated_class, prob_mean in zip(
                    self.all_classes, self.probability_mean
                )
            ],
            axis="columns",
        )
        probabilities = probabilities.div(probabilities.sum(axis=1), axis=0)

        return pd.concat([predictions, probabilities], copy=False, axis="columns")

    fit = fit_noop
    predict_in_sample = predict
    update = fit


class RandomBinaryClassifier(Model):
    """
    A random model that mimics the probability distribution of the target seen during fitting.
    """

    properties = Transformation.Properties(requires_X=False)
    name = "RandomBinaryClassifier"

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        return generate_synthetic_predictions_binary(
            self.memory_target, self.memory_sample_weights, X.index
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        self.memory_target = y
        self.memory_sample_weights = (
            sample_weights
            if sample_weights is not None
            else pd.Series(1, index=y.index)
        )

    predict_in_sample = predict
    update = fit_noop


def generate_synthetic_predictions_binary(
    target: pd.Series,
    sample_weights: pd.Series,
    index: pd.Index,
) -> pd.DataFrame:
    target = target.copy()
    target[target == 0.0] = -1
    prob_mean_class_1 = (target * sample_weights).mean() / 2 + 0.5
    prob_class_1 = np.random.normal(prob_mean_class_1, 0.1, len(index)).clip(0, 1)
    prob_class_0 = 1 - prob_class_1
    return pd.DataFrame(
        {
            "predictions_RandomClassifier": (prob_class_1 > prob_mean_class_1).astype(
                "int"
            ),
            "probabilities_RandomClassifier_0": prob_class_0,
            "probabilities_RandomClassifier_1": prob_class_1,
        },
        index=index,
    )



from typing import Callable, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.core.indexers.objects import BaseIndexer


class FixedForwardWindowIndexerNoTruncation(BaseIndexer):
    """
    Creates window boundaries for fixed-length windows that include the current row.

    Examples
    --------
    >>> df = pd.DataFrame({'B': [0, 1, 2, np.nan, 4]})
    >>> df
         B
    0  0.0
    1  1.0
    2  2.0
    3  NaN
    4  4.0

    >>> indexer = pd.api.indexers.FixedForwardWindowIndexer(window_size=2)
    >>> df.rolling(window=indexer, min_periods=1).sum()
         B
    0  1.0
    1  3.0
    2  2.0
    3  4.0
    4  4.0
    """

    def get_window_bounds(
        self,
        num_values: int = 0,
        min_periods: Optional[int] = None,
        center: Optional[bool] = None,
        closed: Optional[str] = None,
        step: Optional[int] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        if center:
            raise ValueError("Forward-looking windows can't have center=True")
        if closed is not None:
            raise ValueError(
                "Forward-looking windows don't support setting the closed argument"
            )
        if step is None:
            step = 1

        start = np.arange(0, num_values, step, dtype="int64")
        end = start + self.window_size

        return start, end


def create_forward_rolling(
    transformation_func: Optional[Callable],
    agg_func: Callable,
    series: pd.Series,
    period: int,
    shift_by: Optional[int],
    truncate_end: bool,
) -> pd.Series:
    assert period > 0
    indexer = (
        pd.api.indexers.FixedForwardWindowIndexer(window_size=period)
        if truncate_end
        else FixedForwardWindowIndexerNoTruncation(window_size=period)
    )
    shift_by = shift_by if shift_by is not None else -1
    assert shift_by < 0
    transformation_func = transformation_func if transformation_func else lambda x: x
    return agg_func(
        transformation_func(series).rolling(window=indexer, min_periods=1)
    ).shift(shift_by)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import List, Optional, Union

import pandas as pd

from ..base import Transformation, Tunable, fit_noop
from .base import Model


class DummyClassifier(Model, Tunable):
    """
    A model that predicts a predefined class with predefined probabilities.

    Parameters
    ----------
    predicted_value : float, int
        The class to predict.
    all_classes : List[int]
        All possible classes.
    predicted_probabilities : List[float]
        The probabilities returned.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.models import DummyClassifier
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = DummyClassifier(1, [0, 1], [0.5, 0.5])
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                         predictions_DummyClassifier  ...  probabilities_DummyClassifier_1
    2021-12-31 15:40:00                            1  ...                              0.5
    2021-12-31 15:41:00                            1  ...                              0.5
    2021-12-31 15:42:00                            1  ...                              0.5
    2021-12-31 15:43:00                            1  ...                              0.5
    2021-12-31 15:44:00                            1  ...                              0.5
    <BLANKLINE>
    [5 rows x 3 columns]

    ```
    """

    def __init__(
        self,
        predicted_value: Union[float, int],
        all_classes: List[int],
        predicted_probabilities: List[float],
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.predicted_value = predicted_value
        self.all_classes = all_classes
        self.predicted_probabilities = predicted_probabilities
        self.params_to_try = params_to_try
        self.name = name or f"DummyClassifier-{str(self.predicted_value)}"
        self.properties = Transformation.Properties(requires_X=False)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        predictions = pd.Series(
            [self.predicted_value] * len(X),
            index=X.index,
            name="predictions_DummyClassifier",
        )
        probabilities = [
            pd.Series(
                [prob] * len(X),
                index=X.index,
                name=f"probabilities_DummyClassifier_{associated_class}",
            )
            for associated_class, prob in zip(
                self.all_classes, self.predicted_probabilities
            )
        ]

        return pd.concat([predictions] + probabilities, axis="columns")

    predict_in_sample = predict
    fit = fit_noop
    update = fit


class DummyRegressor(Model, Tunable):
    """
    A model that predicts a predefined value.

    Parameters
    ----------
    predicted_value : float
        The value to predict.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.models import DummyRegressor
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = DummyRegressor(0.1)
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                         predictions_DummyRegressor-0.1
    2021-12-31 15:40:00                             0.1
    2021-12-31 15:41:00                             0.1
    2021-12-31 15:42:00                             0.1
    2021-12-31 15:43:00                             0.1
    2021-12-31 15:44:00                             0.1

    ```
    """

    def __init__(
        self,
        predicted_value: float,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.predicted_value = predicted_value
        self.params_to_try = params_to_try
        self.name = name or f"DummyRegressor-{str(self.predicted_value)}"
        self.properties = Transformation.Properties(requires_X=False)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        return pd.Series(
            [self.predicted_value] * len(X),
            index=X.index,
        )

    predict_in_sample = predict
    fit = fit_noop
    update = fit



import pandas as pd

from ...base import EventFilter


class EveryNth(EventFilter):
    def __init__(self, n: int):
        self.n = n

    def get_event_start_times(self, y: pd.Series) -> pd.DatetimeIndex:
        return y.index[:: self.n]



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional, Tuple, Union

import pandas as pd

from fold.base.classes import Artifact

from ..base import (
    Composite,
    InvertibleTransformation,
    Pipeline,
    Pipelines,
    T,
    get_concatenated_names,
)
from ..utils.checks import get_prediction_column, get_prediction_column_name
from ..utils.dataframe import to_series
from ..utils.list import wrap_in_double_list_if_needed


class TransformTarget(Composite):
    """
    Transforms the target within the context of the wrapped Pipeline.
    `wrapped_pipeline` will be applied to the input data, where the target (`y`) is already transformed.
    `y_pipeline` will be applied to the target column.

    The inverse of `y_transformation` will be applied to the predictions of the primary pipeline.

    Eg.: Log or Difference transformation.

    Parameters
    ----------
    wrapped_pipeline: Pipeline
        Pipeline, which will be applied to the input data, where the target (`y`) is already transformed.
    y_pipeline: Union[List[InvertibleTransformation], InvertibleTransformation]
        InvertibleTransformations, which will be applied to the target (`y`)
    invert_wrapped_output: bool, default=True
        Apply the inverse transformation of `y_pipeline` to the output of `wrapped_pipeline`. default is `True`.


    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.composites import ModelResiduals
    >>> from sklearn.linear_model import LinearRegression
    >>> from fold.transformations import Difference
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = TransformTarget(
    ...     wrapped_pipeline=LinearRegression(),
    ...     y_pipeline=Difference(),
    ... )
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)

    ```
    """

    properties = Composite.Properties(
        primary_only_single_pipeline=True,
        secondary_only_single_pipeline=True,
    )

    def __init__(
        self,
        wrapped_pipeline: Pipeline,
        y_pipeline: Union[List[InvertibleTransformation], InvertibleTransformation],
        invert_wrapped_output: bool = True,
        name: Optional[str] = None,
    ) -> None:
        self.wrapped_pipeline = wrap_in_double_list_if_needed(wrapped_pipeline)
        self.y_pipeline = wrap_in_double_list_if_needed(y_pipeline)
        self.invert_wrapped_output = invert_wrapped_output
        self.name = name or "TransformTarget-" + get_concatenated_names(
            self.wrapped_pipeline + self.y_pipeline
        )
        self.metadata = None

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, artifact: Artifact, fit: bool
    ) -> Tuple[pd.DataFrame, T, Artifact]:
        # TransformTarget's primary transformation transforms `y`, not `X`.
        if y is None:
            return (
                pd.DataFrame(),
                None,
                artifact,
            )  # at inference time, `y` will be None, and we don't need to use primary transformations at all, so we return a dummy DataFrame.
        else:
            return y.to_frame(), None, artifact

    def preprocess_secondary(
        self,
        X: pd.DataFrame,
        y: T,
        artifact: Artifact,
        results_primary: pd.DataFrame,
        index: int,
        fit: bool,
    ) -> Tuple[pd.DataFrame, T, Artifact]:
        return X, to_series(results_primary), artifact

    def postprocess_result_primary(
        self,
        results: List[pd.DataFrame],
        y: Optional[pd.Series],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        return results[0]

    def postprocess_result_secondary(
        self,
        primary_results: pd.DataFrame,
        secondary_results: List[pd.DataFrame],
        y: Optional[pd.Series],
        in_sample: bool,
    ) -> pd.DataFrame:
        if self.invert_wrapped_output is False:
            return secondary_results[0]
        predictions = get_prediction_column(secondary_results[0])
        for transformation in reversed(self.y_pipeline[0]):
            predictions = transformation.inverse_transform(predictions, in_sample)
        results = secondary_results[0]
        results[get_prediction_column_name(results)] = to_series(predictions)
        return results

    def get_children_primary(self) -> Pipelines:
        return self.y_pipeline

    def get_children_secondary(
        self,
    ) -> Optional[Pipelines]:
        return self.wrapped_pipeline

    def clone(self, clone_children: Callable) -> TransformTarget:
        clone = TransformTarget(
            wrapped_pipeline=clone_children(self.wrapped_pipeline),
            y_pipeline=clone_children(self.y_pipeline),
            invert_wrapped_output=self.invert_wrapped_output,
        )
        clone.properties = self.properties
        clone.name = self.name
        clone.metadata = self.metadata
        clone.id = self.id
        return clone



import pandas as pd

from ...base import WeightingStrategy
from .max import *  # noqa


class NoWeighting(WeightingStrategy):
    def calculate(self, series: pd.Series) -> pd.Series:
        return pd.Series(1.0, index=series.index)

    def __str__(self) -> str:
        return "NoWeighting"



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Callable, List, Optional

import pandas as pd
import ray
from tqdm import tqdm

from ...base import Artifact, Composite, Pipeline, TrainedPipeline, X
from ...splitters import Fold
from ..types import Backend, Stage


def train_pipeline(
    self,
    func: Callable,
    pipeline: Pipeline,
    X: X,
    y: pd.Series,
    artifact: Artifact,
    splits: List[Fold],
    never_update: bool,
    backend: Backend,
    project_name: str,
    silent: bool,
):
    func = ray.remote(func).options(num_cpus=self.limit_threads)
    X = ray.put(X)
    y = ray.put(y)
    futures = [
        func.remote(
            X,
            y,
            artifact,
            pipeline,
            split,
            never_update,
            backend,
            project_name,
        )
        for split in splits
    ]
    return_value = ray.get(futures)
    ray.internal.free([X, y])
    return return_value


def backtest_pipeline(
    self,
    func: Callable,
    pipeline: TrainedPipeline,
    splits: List[Fold],
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    backend: Backend,
    mutate: bool,
    silent: bool,
):
    func = ray.remote(func).options(num_cpus=self.limit_threads)
    X = ray.put(X)
    y = ray.put(y)
    pipeline = ray.put(pipeline)

    # result_refs = []
    # for split in splits:
    #     if len(result_refs) > 1:  # self.limit_threads:
    #         ready_refs, result_refs = ray.wait(result_refs, num_returns=1)
    #         ray.get(ready_refs)

    #     result_refs.append(
    #         func.remote(pipeline, split, X, y, artifact, backend, mutate),
    #     )

    # return_value_1 = sorted(ray.get(result_refs), key=lambda x: x[0].index[0])

    futures = [
        func.remote(pipeline, split, X, y, artifact, backend, mutate)
        for split in splits
    ]
    return_value = ray.get(futures)

    ray.internal.free([X, y, pipeline])
    return return_value


def process_child_transformations(
    self,
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: X,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
):
    # list_of_child_transformations_with_index = list(
    #     list_of_child_transformations_with_index
    # )
    # available_resources = ray.available_resources()
    # if (
    #     len(list_of_child_transformations_with_index) == 1
    #     or "CPU" not in available_resources
    #     or ("CPU" in available_resources and available_resources["CPU"] < 2)
    # ):
    return [
        func(
            composite,
            index,
            child_transformation,
            X,
            y,
            artifacts,
            stage,
            backend,
            results_primary,
            tqdm,
        )
        for index, child_transformation in list_of_child_transformations_with_index
    ]
    # else:
    #     func = ray.remote(func)
    #     X = ray.put(X)
    #     y = ray.put(y)

    #     futures = [
    #         func.remote(
    #             composite,
    #             index,
    #             child_transformation,
    #             X,
    #             y,
    #             artifacts,
    #             stage,
    #             backend,
    #             results_primary,
    #         )
    #         for index, child_transformation in list_of_child_transformations_with_index
    #     ]
    # batch_size = 2

    # chunk processing of futures with ray
    # results = []
    # for i in range(0, len(futures), batch_size):
    #     results.extend(ray.get(futures[i : i + batch_size]))
    # while futures:
    #     if len(futures) < batch_size:
    #         batch_size = len(futures)
    #     ready_futures, futures = ray.wait(futures, num_returns=batch_size)
    #     result = ray.get(ready_futures)
    #     results.append(result)

    # return results


class RayBackend(Backend):
    name = "ray"
    process_child_transformations = process_child_transformations
    train_pipeline = train_pipeline
    backtest_pipeline = backtest_pipeline

    def __init__(self, limit_threads: Optional[int] = None):
        self.limit_threads = limit_threads



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Optional, Tuple, TypeVar

import pandas as pd

from ..base import Transformation
from ..utils.dataframe import concat_on_index

T = TypeVar("T", pd.Series, Optional[pd.Series])
S = TypeVar("S", pd.Series, Optional[pd.Series])


def preprocess_X_y_with_memory(
    transformation: Transformation,
    X: pd.DataFrame,
    y: T,
    sample_weights: S,
    in_sample: bool,
) -> Tuple[pd.DataFrame, T, S]:
    if transformation.properties.disable_memory:
        return X, y, sample_weights
    memory_size = transformation.properties.memory_size
    if (
        not hasattr(transformation, "_state")
        or transformation._state is None
        or memory_size is None
    ):
        return X, y, sample_weights
    memory_X, memory_y, memory_sample_weights = (
        transformation._state.memory_X,
        transformation._state.memory_y,
        transformation._state.memory_sample_weights,
    )
    non_overlapping_indices = ~memory_X.index.isin(X.index)
    memory_X, memory_y, memory_sample_weights = (
        memory_X[non_overlapping_indices],
        memory_y[non_overlapping_indices],
        memory_sample_weights[non_overlapping_indices]
        if memory_sample_weights is not None
        else None,
    )
    if len(memory_X) == 0:
        return X, y, sample_weights
    assert len(memory_X) == len(memory_y)
    if memory_sample_weights is not None:
        assert len(memory_X) == len(memory_sample_weights)
    if y is None:
        assert sample_weights is None
        return (
            concat_on_index([memory_X.iloc[-memory_size:None], X], copy=True),
            y,
            sample_weights,
        )
    elif in_sample is True:
        memory_y.name = y.name
        return (
            concat_on_index([memory_X, X], copy=True),
            concat_on_index([memory_y, y], copy=True),
            concat_on_index([memory_sample_weights, sample_weights], copy=True),
        )
    else:
        memory_y.name = y.name
        return (
            concat_on_index([memory_X.iloc[-memory_size:None], X], copy=True),
            concat_on_index(
                [memory_y.iloc[-memory_size:None], y],
                copy=True,
            ),
            concat_on_index(
                [memory_sample_weights.iloc[-memory_size:None], sample_weights],
                copy=True,
            )
            if sample_weights is not None
            else None,
        )


def postprocess_X_y_into_memory_(
    transformation: Transformation,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series],
    in_sample: bool,
) -> None:
    if transformation.properties.disable_memory:
        return
    # This function mutates the transformation's state property
    # don't update the transformation if we're in inference mode (y is None)
    memory_size = transformation.properties.memory_size
    if memory_size is None or y is None:
        return

    window_size = memory_size
    if in_sample:
        # store the whole training X and y
        transformation._state = Transformation.State(
            memory_X=X, memory_y=y, memory_sample_weights=sample_weights
        )
    elif memory_size < len(X):
        memory_X, memory_y, memory_sample_weights = (
            transformation._state.memory_X,
            transformation._state.memory_y,
            transformation._state.memory_sample_weights,
        )
        memory_y.name = y.name
        #  memory requirement is greater than the current batch, so we use the previous memory as well
        transformation._state = Transformation.State(
            memory_X=concat_on_index([memory_X, X], copy=True).iloc[-window_size:None],
            memory_y=concat_on_index([memory_y, y], copy=True).iloc[-window_size:None],
            memory_sample_weights=concat_on_index(
                [memory_sample_weights, sample_weights], copy=True
            ).iloc[-window_size:None]
            if sample_weights is not None
            else None,
        )
    else:
        transformation._state = Transformation.State(
            memory_X=X.iloc[-window_size:None],
            memory_y=y.iloc[-window_size:None],
            memory_sample_weights=sample_weights.iloc[-window_size:None]
            if sample_weights is not None
            else None,
        )



from typing import Union

import pandas as pd


def calculate_rolling_window_size(
    window_size: Union[int, float], series: pd.Series
) -> int:
    return window_size if window_size > 1 else int(len(series) * window_size)  # type: ignore



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional, Union

import pandas as pd

from ..base import Artifact, Composite, Pipelines, Transformation, Tunable
from ..utils.list import wrap_in_list
from .utils import _check_for_duplicate_names


class SelectBest(Composite, Tunable):
    selected_: Optional[str] = None

    def __init__(
        self,
        choose_from: Pipelines,
        name: Optional[str] = None,
    ) -> None:
        self.choose_from = choose_from
        _check_for_duplicate_names(self.choose_from)
        self.name = name or "SelectBest"
        self.properties = Composite.Properties()
        self.metadata = None

    @classmethod
    def from_cloned_instance(
        cls,
        choose_from: Pipelines,
        selected_: Optional[str],
        name: Optional[str],
        metadata: Optional[Composite.Metadata],
        id: str,
    ) -> SelectBest:
        instance = cls(choose_from)
        instance.selected_ = selected_
        instance.name = name
        instance.metadata = metadata
        instance.id = id
        return instance

    def postprocess_result_primary(
        self,
        results: List[pd.DataFrame],
        y: Optional[pd.Series],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        assert self.selected_ is not None, ValueError(
            "SelectBest only works within an `Optimize` class."
        )
        return results[0]

    def get_children_primary(self) -> Pipelines:
        selected = get_candidate_by_name(self.choose_from, self.selected_)
        if selected is None:
            return self.choose_from
        else:
            return wrap_in_list(selected)

    def clone(self, clone_children: Callable) -> SelectBest:
        return SelectBest.from_cloned_instance(
            choose_from=clone_children(self.choose_from),
            selected_=self.selected_,
            name=self.name,
            metadata=self.metadata,
            id=self.id,
        )

    def get_params(self) -> dict:
        return dict(
            selected_=self.selected_,
            name=self.name,
        )

    def get_params_to_try(self) -> Optional[dict]:
        return {"selected_": [i.name for i in self.choose_from]}

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        assert clone_children is not None
        # This is a bit hacky, we "peak" into the new params, and check if selected_ has been set,
        # if yes, we already only clone the selected candidate, not all, to enable conditional hyperparameter optimization.
        # This is our way of filtering out objects that should not be tuned during this round.
        if "selected_" in parameters:
            children = (
                self.choose_from
                if parameters["selected_"] is None
                else [get_candidate_by_name(self.choose_from, parameters["selected_"])]
            )
        else:
            children = self.choose_from
        return SelectBest.from_cloned_instance(
            choose_from=clone_children(children),
            selected_=parameters["selected_"],
            name=self.name,
            metadata=self.metadata,
            id=self.id,
        )


def get_candidate_by_name(
    candidates: List[Union[Transformation, Composite]], name: str
) -> Optional[Union[Transformation, Composite]]:
    for candidate in candidates:
        if candidate.name == name:
            return candidate
    return None



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional

import pandas as pd

from fold.utils.checks import get_probabilities_columns, has_probabilities

from ..base import Artifact, Composite, Pipelines, get_concatenated_names
from .columns import average_results


class Ensemble(Composite):
    """
    Ensemble (average) the results of multiple pipelines.

    Parameters
    ----------

    pipelines : Pipelines
        A list of pipelines to be applied to the data, independently of each other.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import Ensemble
        >>> from fold.models import DummyRegressor
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
        >>> pipeline = Ensemble([
        ...     DummyRegressor(0.1),
        ...     DummyRegressor(0.9),
        ... ])
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
        >>> preds.squeeze().head()
        2021-12-31 15:40:00    0.5
        2021-12-31 15:41:00    0.5
        2021-12-31 15:42:00    0.5
        2021-12-31 15:43:00    0.5
        2021-12-31 15:44:00    0.5
        Freq: T, Name: predictions_Ensemble-DummyRegressor-0.1-DummyRegressor-0.9, dtype: float64
    """

    def __init__(
        self,
        pipelines: Pipelines,
        reconstruct_predictions_from_probabilities: bool = False,
        verbose: bool = False,
        name: Optional[str] = None,
    ) -> None:
        self.pipelines = pipelines
        self.name = name or "Ensemble-" + get_concatenated_names(pipelines)
        self.reconstruct_predictions_from_probabilities = (
            reconstruct_predictions_from_probabilities
        )
        self.verbose = verbose
        self.properties = Composite.Properties()
        self.metadata = None

    def postprocess_result_primary(
        self,
        results: List[pd.DataFrame],
        y: Optional[pd.Series],
        original_artifact: Artifact,
        fit: bool,
    ) -> pd.DataFrame:
        if self.verbose and all([has_probabilities(result) for result in results]):
            probs = pd.concat(
                [get_probabilities_columns(result).iloc[:, 0] for result in results],
                axis="columns",
            )
            print(
                f"Ensemble - Avg correlation of probabilities: {probs.corr().mean().mean()}"
            )
        return average_results(
            results, self.name, self.reconstruct_predictions_from_probabilities
        )

    def get_children_primary(self) -> Pipelines:
        return self.pipelines

    def clone(self, clone_children: Callable) -> Ensemble:
        clone = Ensemble(
            pipelines=clone_children(self.pipelines),
            reconstruct_predictions_from_probabilities=self.reconstruct_predictions_from_probabilities,
            verbose=self.verbose,
        )
        clone.properties = self.properties
        clone.name = self.name
        clone.metadata = self.metadata
        clone.id = self.id
        return clone



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional, Union

import pandas as pd

from ..base import SingleFunctionTransformation, Transformation, Tunable, fit_noop
from ..utils.enums import ParsableEnum
from ..utils.list import wrap_in_list


class DateTimeFeature(ParsableEnum):
    second = "second"
    minute = "minute"
    hour = "hour"
    day_of_week = "day_of_week"
    day_of_month = "day_of_month"
    day_of_year = "day_of_year"
    week = "week"
    week_of_year = "week_of_year"
    month = "month"
    quarter = "quarter"
    year = "year"


class AddDateTimeFeatures(Transformation, Tunable):
    """
    Adds (potentially multiple) date/time features to the input, as additional columns.
    The name of the new column will be the name of the DateTimeFeature passed in.
    Values are returned as integers, so the 59th minute of the hour will be `59`, and year 2022 will be `2022`.

    Parameters
    ----------

    features: List[DateTimeFeature, str]
        The features to add to the input. Options: `second`, `minute`, `hour`, `day_of_week`, `day_of_month`, `day_of_year`, `week`, `week_of_year`, `month`, `quarter`, `year`.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddDateTimeFeatures
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddDateTimeFeatures(["minute"])
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  minute
    2021-12-31 15:40:00 -0.0000      40
    2021-12-31 15:41:00  0.0126      41
    2021-12-31 15:42:00  0.0251      42
    2021-12-31 15:43:00  0.0377      43
    2021-12-31 15:44:00  0.0502      44

    ```
    """

    def __init__(
        self,
        features: List[Union[DateTimeFeature, str]],
        keep_original: bool = True,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.features = [DateTimeFeature(f) for f in wrap_in_list(features)]
        self.name = (
            name
            if name is not None
            else f"AddDateTimeFeatures-{'-'.join([i.value for i in self.features])}"
        )
        self.keep_original = keep_original
        self.params_to_try = params_to_try
        self.properties = Transformation.Properties(requires_X=False)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        X_datetime = pd.DataFrame([], index=X.index)
        for feature in self.features:
            if feature == DateTimeFeature.second:
                X_datetime[feature.value] = X.index.second
            elif feature == DateTimeFeature.minute:
                X_datetime[feature.value] = X.index.minute
            elif feature == DateTimeFeature.hour:
                X_datetime[feature.value] = X.index.hour
            elif feature == DateTimeFeature.day_of_week:
                X_datetime[feature.value] = X.index.dayofweek
            elif feature == DateTimeFeature.day_of_month:
                X_datetime[feature.value] = X.index.day
            elif feature == DateTimeFeature.day_of_year:
                X_datetime[feature.value] = X.index.dayofyear
            elif (
                feature == DateTimeFeature.week
                or feature == DateTimeFeature.week_of_year
            ):
                X_datetime[feature.value] = pd.Index(
                    X.index.isocalendar().week, dtype="int"
                )
            elif feature == DateTimeFeature.month:
                X_datetime[feature.value] = X.index.month
            elif feature == DateTimeFeature.quarter:
                X_datetime[feature.value] = X.index.quarter
            elif feature == DateTimeFeature.year:
                X_datetime[feature.value] = X.index.year
            else:
                raise ValueError(f"Unsupported feature: {feature}")
        to_concat = [X, X_datetime] if self.keep_original else X_datetime
        concatenated = pd.concat(to_concat, copy=False, axis="columns")
        return concatenated

    fit = fit_noop
    update = fit_noop


class AddSecond(SingleFunctionTransformation):
    """
    Adds "second" features to the input, as an additional column.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddSecond
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="S")
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddSecond()
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  second
    2021-12-31 23:51:40 -0.0000      40
    2021-12-31 23:51:41  0.0126      41
    2021-12-31 23:51:42  0.0251      42
    2021-12-31 23:51:43  0.0377      43
    2021-12-31 23:51:44  0.0502      44

    ```
    """

    name = "AddSecond"

    def get_function(self) -> Callable:
        return lambda X: X.index.second.to_series().rename("second").set_axis(X.index)


class AddMinute(SingleFunctionTransformation):
    """
    Adds "minute" features to the input, as an additional column.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddMinute
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddMinute()
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  minute
    2021-12-31 15:40:00 -0.0000      40
    2021-12-31 15:41:00  0.0126      41
    2021-12-31 15:42:00  0.0251      42
    2021-12-31 15:43:00  0.0377      43
    2021-12-31 15:44:00  0.0502      44

    ```
    """

    name = "AddMinute"

    def get_function(self) -> Callable:
        return lambda X: X.index.minute.to_series().rename("minute").set_axis(X.index)


class AddHour(SingleFunctionTransformation):
    name = "AddHour"

    def get_function(self) -> Callable:
        return lambda X: X.index.hour.to_series().rename("hour").set_axis(X.index)


class AddDayOfWeek(SingleFunctionTransformation):
    name = "AddDayOfWeek"

    def get_function(self) -> Callable:
        return (
            lambda X: X.index.dayofweek.to_series()
            .rename("day_of_week")
            .set_axis(X.index)
        )


class AddDayOfMonth(SingleFunctionTransformation):
    name = "AddDayOfMonth"

    def get_function(self) -> Callable:
        return (
            lambda X: X.index.day.to_series().rename("day_of_month").set_axis(X.index)
        )


class AddDayOfYear(SingleFunctionTransformation):
    name = "AddDayOfYear"

    def get_function(self) -> Callable:
        return (
            lambda X: X.index.dayofyear.to_series()
            .rename("day_of_year")
            .set_axis(X.index)
        )


class AddWeek(SingleFunctionTransformation):
    name = "AddWeek"

    def get_function(self) -> Callable:
        return (
            lambda X: pd.Index(X.index.isocalendar().week, dtype="int")
            .to_series()
            .set_axis(X.index)
            .rename("week")
        )


class AddWeekOfYear(SingleFunctionTransformation):
    name = "AddWeekOfYear"

    def get_function(self) -> Callable:
        return (
            lambda X: pd.Index(X.index.isocalendar().week, dtype="int")
            .to_series()
            .set_axis(X.index)
            .rename("week_of_year")
        )


class AddMonth(SingleFunctionTransformation):
    name = "AddMonth"

    def get_function(self) -> Callable:
        return lambda X: X.index.month.to_series().rename("month").set_axis(X.index)


class AddQuarter(SingleFunctionTransformation):
    name = "AddQuarter"

    def get_function(self) -> Callable:
        return lambda X: X.index.quarter.to_series().rename("quarter").set_axis(X.index)


class AddYear(SingleFunctionTransformation):
    name = "AddYear"

    def get_function(self) -> Callable:
        return lambda X: X.index.year.to_series().rename("year").set_axis(X.index)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from .base import *
from .baseline import *
from .dummy import *
from .random import *
from .sklearn import *



from typing import Optional, Union

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression, SGDRegressor

from fold.base import Tunable
from fold.models.base import TimeSeriesModel


class AR(TimeSeriesModel, Tunable):
    def __init__(self, p: int, params_to_try: Optional[dict] = None) -> None:
        self.p = p
        self.name = f"AR-{str(p)}"
        self.properties = TimeSeriesModel.Properties(
            requires_X=False,
            mode=TimeSeriesModel.Properties.Mode.online,
            model_type=TimeSeriesModel.Properties.ModelType.regressor,
            memory_size=p,
            _internal_supports_minibatch_backtesting=True,
        )
        self.models = [LinearRegression() for _ in range(p)]
        self.params_to_try = params_to_try

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        # Using Least Squares as it's faster than SGD for the initial fit
        for index, model in enumerate(self.models, start=1):
            model.fit(
                y.shift(index).to_frame()[index:],
                y[index:],
                sample_weight=sample_weights[index:]
                if sample_weights is not None
                else None,
            )
        self.parameters = [
            {
                "coef_": model.coef_[0],
                "intercept_": model.intercept_,
            }
            for model in self.models
        ]

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if isinstance(self.models[0], LinearRegression):
            # If the model is being updated for the first time, convert to SGDRegressor (with the parameters from the initial fit)
            self.models = [SGDRegressor(warm_start=True) for _ in range(self.p)]
            for index, (model, parameters) in enumerate(
                zip(self.models, self.parameters), start=1
            ):
                model.fit(
                    y.shift(index).to_frame()[index:],
                    y[index:],
                    coef_init=parameters["coef_"],
                    intercept_init=parameters["intercept_"],
                    sample_weight=sample_weights[index:]
                    if sample_weights is not None
                    else None,
                )
        else:
            # For any subsequent updates, just call partial_fit
            for index, model in enumerate(self.models, start=1):
                model.partial_fit(
                    y.shift(index).to_frame()[index:],
                    y[index:],
                    sample_weight=sample_weights[index:]
                    if sample_weights is not None
                    else None,
                )

    def predict(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        return _predict(self.models, past_y, indices=X.index).fillna(0.0)

    predict_in_sample = predict

    def get_params(self) -> dict:
        return {"p": self.p}


def _predict(models, past_y: pd.Series, indices) -> pd.Series:
    preds = [
        np.concatenate(
            [
                np.zeros((index,)),
                lr.predict(past_y.shift(index - 1).to_frame()[index:]),
            ]
        )
        for index, lr in enumerate(models, start=1)
    ]
    return pd.Series(np.vstack(preds).mean(axis=0), index=indices)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

from __future__ import annotations

from typing import Callable, List, Optional

import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm

from ...base import Artifact, Composite, Pipeline, TrainedPipeline, X
from ...splitters import Fold
from ..types import Backend, Stage


def train_pipeline(
    self: JoblibBackend,
    func: Callable,
    pipeline: Pipeline,
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    splits: List[Fold],
    never_update: bool,
    backend: Backend,
    project_name: str,
    silent: bool,
):
    return Parallel(
        n_jobs=self.limit_threads, prefer="threads" if self.prefer_threads else None
    )(
        [
            delayed(func)(
                X, y, artifact, pipeline, split, never_update, backend, project_name
            )
            for split in splits
        ]
    )


def backtest_pipeline(
    self: JoblibBackend,
    func: Callable,
    pipeline: TrainedPipeline,
    splits: List[Fold],
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    backend: Backend,
    mutate: bool,
    silent: bool,
):
    return Parallel(
        n_jobs=self.limit_threads, prefer="threads" if self.prefer_threads else None
    )(
        [
            delayed(func)(pipeline, split, X, y, artifact, backend, mutate)
            for split in splits
        ]
    )


def process_child_transformations(
    self: JoblibBackend,
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: X,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
):
    # return Parallel(
    #     n_jobs=self.limit_threads, prefer="threads" if self.prefer_threads else None
    # )(
    #     [
    #         delayed(func)(
    #             composite,
    #             index,
    #             child_transformation,
    #             X,
    #             y,
    #             artifacts,
    #             stage,
    #             backend,
    #             results_primary,
    #             None,
    #         )
    #         for index, child_transformation in list_of_child_transformations_with_index
    #     ]
    # )
    return [
        func(
            composite,
            index,
            child_transformation,
            X,
            y,
            artifacts,
            stage,
            backend,
            results_primary,
            tqdm,
        )
        for index, child_transformation in list_of_child_transformations_with_index
    ]


class JoblibBackend(Backend):
    name = "joblib"
    process_child_transformations = process_child_transformations
    train_pipeline = train_pipeline
    backtest_pipeline = backtest_pipeline

    def __init__(
        self, limit_threads: Optional[int] = None, prefer_threads: bool = True
    ):
        self.limit_threads = limit_threads
        self.prefer_threads = prefer_threads



from __future__ import annotations

from abc import abstractmethod
from math import sqrt
from typing import Any, Callable, Optional, Type, Union

import pandas as pd

from fold.base import Tunable
from fold.models.base import Model

from .types import ClassWeightingStrategy


class WrapGBM(Model, Tunable):
    def __init__(
        self,
        model_class: Type,
        init_args: Optional[dict] = {},
        instance: Optional[Any] = None,
        set_class_weights: Union[
            ClassWeightingStrategy, str
        ] = ClassWeightingStrategy.none,
        params_to_try: Optional[dict] = None,
        name: Optional[str] = None,
    ) -> None:
        self.init_args = init_args
        self.model_class = model_class

        self.model = model_class(**init_args) if instance is None else instance
        self.set_class_weights = ClassWeightingStrategy.from_str(set_class_weights)

        self.properties = Model.Properties(
            requires_X=True, model_type=self.get_model_type(self.model)
        )

        self.name = name or self.model.__class__.__name__
        self.params_to_try = params_to_try

    @abstractmethod
    def get_model_type(self, model) -> Model.Properties.ModelType:
        raise NotImplementedError

    @classmethod
    def from_model(
        cls,
        model,
        set_class_weights: Union[
            ClassWeightingStrategy, str
        ] = ClassWeightingStrategy.none,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ):
        return cls(
            model.__class__,
            init_args=model.get_params(),
            instance=model,
            set_class_weights=set_class_weights,
            name=name,
            params_to_try=params_to_try,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        if self.set_class_weights in [
            ClassWeightingStrategy.balanced,
            ClassWeightingStrategy.balanced_sqrt,
        ]:
            counts = y.value_counts()
            scale_pos_weight = counts[0] / counts[1]
            if self.set_class_weights == ClassWeightingStrategy.balanced_sqrt:
                scale_pos_weight = sqrt(scale_pos_weight)
            self.model = self.model.set_params(
                **dict(scale_pos_weight=scale_pos_weight)
            )
            self.model.fit(
                X=X,
                y=y,
                sample_weight=sample_weights,
            )
        else:
            self.model.fit(X=X, y=y, sample_weight=sample_weights)

    def update(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        raise NotImplementedError

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        predictions = pd.Series(self.model.predict(X), index=X.index).rename(
            f"predictions_{self.name}"
        )
        if self.properties.model_type == Model.Properties.ModelType.classifier:
            probabilities = pd.DataFrame(
                data=self.model.predict_proba(X),
                index=X.index,
                columns=[
                    f"probabilities_{self.name}_{float(item)}"
                    for item in self.model.classes_
                ],
            )
            return pd.concat([predictions, probabilities], axis="columns")
        else:
            return predictions

    predict_in_sample = predict

    def get_params(self) -> dict:
        return {
            **self.model.get_params(),
            **dict(set_class_weights=self.set_class_weights.value),
        }

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        if "set_class_weights" in parameters:
            set_class_weights = parameters.pop("set_class_weights")
        return self.__class__(
            self.model_class,
            init_args=parameters,
            set_class_weights=set_class_weights,
            name=self.name,
        )


class WrapLGBM(WrapGBM):
    def get_model_type(self, model) -> Model.Properties.ModelType:
        from lightgbm import LGBMClassifier, LGBMRegressor

        if isinstance(model, LGBMRegressor):
            return Model.Properties.ModelType.regressor
        elif isinstance(model, LGBMClassifier):
            return Model.Properties.ModelType.classifier
        else:
            raise ValueError(f"Unknown model type: {type(model)}")

    def update(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        self.model.fit(
            X,
            y,
            sample_weight=sample_weights,
            init_model=self.model.get_booster(),
        )


class WrapXGB(WrapGBM):
    def get_model_type(self, model) -> Model.Properties.ModelType:
        from xgboost import XGBClassifier, XGBRegressor, XGBRFClassifier, XGBRFRegressor

        if isinstance(self.model, XGBRegressor) or isinstance(
            self.model, XGBRFRegressor
        ):
            return Model.Properties.ModelType.regressor
        elif isinstance(self.model, XGBClassifier) or isinstance(
            self.model, XGBRFClassifier
        ):
            return Model.Properties.ModelType.classifier
        else:
            raise ValueError(f"Unknown model type: {type(self.model)}")

    def update(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        self.model.fit(
            X=X,
            y=y,
            xgb_model=self.model.get_booster(),
            sample_weight=sample_weights,
        )



from __future__ import annotations

from typing import Any, Dict, Optional, Type, Union

import pandas as pd

from fold.models.base import Model
from fold.utils.checks import is_X_available


class WrapSktime(Model):
    def __init__(
        self,
        model_class: Type,
        init_args: Optional[Dict],
        use_exogenous: Optional[bool] = None,
        online_mode: bool = False,
        instance: Optional[Any] = None,
    ) -> None:
        self.init_args = init_args
        init_args = {} if init_args is None else init_args
        self.model = model_class(**init_args) if instance is None else instance
        self.model_class = model_class
        self.use_exogenous = use_exogenous
        self.properties = Model.Properties(
            requires_X=False,
            model_type=Model.Properties.ModelType.regressor,
            mode=(
                Model.Properties.Mode.online
                if online_mode
                else Model.Properties.Mode.minibatch
            ),
        )
        self.name = self.model_class.__class__.__name__

    @classmethod
    def from_model(
        cls,
        model,
        use_exogenous: Optional[bool] = None,
        online_mode: bool = False,
    ) -> WrapSktime:
        return cls(
            model_class=None,
            init_args=None,
            use_exogenous=use_exogenous,
            instance=model,
            online_mode=online_mode,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            self.model.fit(y=y, X=X)
        else:
            self.model.fit(y=y)

    def update(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        if not hasattr(self.model, "update"):
            return
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            self.model = self.model.update(y=y, X=X, update_params=True)
        else:
            self.model = self.model.update(y=y, update_params=True)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        from sktime.forecasting.base import ForecastingHorizon

        fh = ForecastingHorizon(list(range(1, len(X) + 1)), is_relative=True)
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            return self.model.predict(fh, X=X)
        else:
            return self.model.predict(fh)

    def predict_in_sample(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        from sktime.forecasting.base import ForecastingHorizon

        fh = ForecastingHorizon(X.index, is_relative=False)
        use_exogenous = (
            is_X_available(X) if self.use_exogenous is None else self.use_exogenous
        )
        if use_exogenous:
            return self.model.predict(fh, X=X)
        else:
            return self.model.predict(fh)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

import pandas as pd

from ..base import (
    Artifact,
    EventDataFrame,
    InSamplePredictions,
    OutOfSamplePredictions,
    Pipeline,
    PipelineCard,
    TrainedPipelineCard,
)
from ..base.scoring import score_results
from ..splitters import Splitter
from ..utils.dataframe import concat_on_index_override_duplicate_rows
from .backtesting import backtest
from .training import train
from .types import Backend, BackendType, TrainMethod

if TYPE_CHECKING:
    from krisi import ScoreCard


def backtest_score(
    trained_pipelines: TrainedPipelineCard,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[BackendType, Backend, str] = BackendType.no,
    events: Optional[EventDataFrame] = None,
    silent: bool = False,
    return_artifacts: bool = False,
    krisi_args: Optional[dict] = None,
) -> Union[
    Tuple["ScoreCard", OutOfSamplePredictions],
    Tuple["ScoreCard", OutOfSamplePredictions, Artifact],
]:
    """
    Run backtest then scoring.
    [`krisi`](https://github.com/dream-faster/krisi) is required to be installed.

    Parameters
    ----------
    trained_pipelines: TrainedPipelineCard
        The fitted pipelines, for all folds.
    X: Optional[pd.DataFrame]
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    events: EventDataFrame, optional = None
        Events that should be passed into the pipeline, by default None.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    return_artifacts: bool = False
        Whether to return the artifacts of the training process, by default False.
    krisi_args: Optional[Dict[str, Any]] = None
        Arguments that will be passed into `krisi` score function, by default None.

    Returns
    -------
    "ScoreCard"
        A ScoreCard from `krisi`.
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    """

    pred, artifacts = backtest(
        trained_pipelines,
        X,
        y,
        splitter,
        backend=backend,
        events=events,
        silent=silent,
        return_artifacts=True,
    )

    scorecard, pred = score_results(
        pred,
        y,
        artifacts=artifacts,
        krisi_args=krisi_args,
    )
    if return_artifacts:
        return scorecard, pred, artifacts
    else:
        return scorecard, pred


def train_backtest(
    pipeline: Union[Pipeline, PipelineCard],
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[BackendType, Backend, str] = BackendType.no,
    events: Optional[EventDataFrame] = None,
    train_method: Union[TrainMethod, str] = TrainMethod.parallel,
    silent: bool = False,
    return_artifacts: bool = False,
    return_insample: bool = False,
) -> Union[
    Tuple[OutOfSamplePredictions, TrainedPipelineCard],
    Tuple[OutOfSamplePredictions, TrainedPipelineCard, Artifact],
    Tuple[OutOfSamplePredictions, TrainedPipelineCard, Artifact, InSamplePredictions],
]:
    """
    Run train and backtest.

    Parameters
    ----------
    pipeline: Union[Pipeline, PipelineCard]
        The pipeline to be fitted.
    X: Optional[pd.DataFrame]
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    events: EventDataFrame, optional = None
        Events that should be passed into the pipeline, by default None.
    train_method: TrainMethod, str = TrainMethod.parallel
        The training methodology, by default `parallel`.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    return_artifacts: bool = False
        Whether to return the artifacts of the process, by default False.
    return_insample: bool = False

    Returns
    -------
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    TrainedPipelineCard
        The fitted pipelines, for all folds.
    """
    trained_pipelines, train_artifacts, insample_predictions = train(
        pipeline,
        X,
        y,
        splitter,
        events=events,
        train_method=train_method,
        backend=backend,
        silent=silent,
        return_artifacts=True,
        return_insample=True,
    )

    pred, backtest_artifacts = backtest(
        trained_pipelines,
        X,
        y,
        splitter,
        backend=backend,
        events=events,
        silent=silent,
        return_artifacts=True,
    )

    if return_artifacts:
        artifacts = concat_on_index_override_duplicate_rows(
            [train_artifacts, backtest_artifacts],
        )
        assert artifacts.index.is_monotonic_increasing
        if return_insample:
            return pred, trained_pipelines, artifacts, insample_predictions
        else:
            return pred, trained_pipelines, artifacts
    else:
        if return_insample:
            return pred, trained_pipelines, insample_predictions
        else:
            return pred, trained_pipelines


def train_evaluate(
    pipeline: Union[Pipeline, PipelineCard],
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[BackendType, Backend, str] = BackendType.no,
    events: Optional[EventDataFrame] = None,
    train_method: Union[TrainMethod, str] = TrainMethod.parallel,
    silent: bool = False,
    return_artifacts: bool = False,
    return_insample_scorecard: bool = False,
    krisi_args: Optional[Dict[str, Any]] = None,
) -> Union[
    Tuple[
        "ScoreCard",
        OutOfSamplePredictions,
        TrainedPipelineCard,
    ],
    Tuple[
        "ScoreCard",
        OutOfSamplePredictions,
        TrainedPipelineCard,
        Artifact,
        "ScoreCard",
    ],
    Tuple[
        "ScoreCard",
        OutOfSamplePredictions,
        TrainedPipelineCard,
        Artifact,
    ],
    Tuple[
        "ScoreCard",
        OutOfSamplePredictions,
        TrainedPipelineCard,
        "ScoreCard",
    ],
]:
    """
    Run train, backtest then run scoring.
    [`krisi`](https://github.com/dream-faster/krisi) needs to be installed.

    Parameters
    ----------
    pipeline: Union[Pipeline, PipelineCard]
        The pipeline to be fitted.
    X: pd.DataFrame, optional
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    events: EventDataFrame, optional = None
        Events that should be passed into the pipeline, by default None.
    train_method: TrainMethod, str = TrainMethod.parallel
        The training methodology, by default `parallel`.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    krisi_args: Dict[str, Any], optional = None
        Arguments that will be passed into `krisi` score function, by default None.

    Returns
    -------
    "ScoreCard"
        A ScoreCard from `krisi`.
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    TrainedPipelineCard
        The fitted pipelines, for all folds.
    """
    trained_pipelines, train_artifacts, insample_predictions = train(
        pipeline,
        X,
        y,
        splitter,
        events=events,
        train_method=train_method,
        backend=backend,
        silent=silent,
        return_artifacts=True,
        return_insample=True,
    )

    scorecard, pred, backtest_artifacts = backtest_score(
        trained_pipelines,
        X,
        y,
        splitter,
        backend=backend,
        events=events,
        silent=silent,
        return_artifacts=True,
        krisi_args=dict(sample_type="validation", **(krisi_args or {})),
    )

    return_object = [scorecard, pred, trained_pipelines]

    if return_artifacts:
        concatinated_artifacts = concat_on_index_override_duplicate_rows(
            [train_artifacts, backtest_artifacts]
        )
        return_object.append(concatinated_artifacts)

    if return_insample_scorecard:
        scorecard_insample, _ = score_results(
            insample_predictions,
            y,
            artifacts=train_artifacts,
            krisi_args=dict(sample_type="insample", **(krisi_args or {})),
        )
        return_object.append(scorecard_insample)

    return tuple(return_object)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

import logging
from typing import Optional, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

from ..base import Artifact

T = TypeVar("T", pd.Series, Optional[pd.Series])
logger = logging.getLogger("fold:utils")


def trim_initial_nans(
    X: pd.DataFrame, y: pd.Series, artifact: Artifact
) -> Tuple[pd.DataFrame, pd.Series, Artifact]:
    # Optimize for speed, if the first value is not NaN, we can save all the subsequent computation
    if not X.iloc[0].isna().any() and (y is None or not np.isnan(y.iloc[0])):
        return X, y, artifact
    first_valid_index_X = get_first_valid_index(X)
    first_valid_index_y = get_first_valid_index(y)
    if first_valid_index_X is None or first_valid_index_y is None:
        return (
            pd.DataFrame(),
            pd.Series(dtype="float64"),
            Artifact.empty(X.index),
        )
    first_valid_index = max(first_valid_index_X, first_valid_index_y)
    if first_valid_index == 0:
        return X, y, artifact
    else:
        logger.warn(
            f"The first {first_valid_index} rows of the dataset were removed because they contained NaN values."
        )
        return (
            X.iloc[first_valid_index:],
            y.iloc[first_valid_index:],
            artifact.iloc[first_valid_index:],
        )


def trim_initial_nans_single(X: pd.DataFrame) -> pd.DataFrame:
    # Optimize for speed, if the first value is not NaN, we can save all the subsequent computation
    if not X.iloc[0].isna().any():
        return X
    first_valid_index = get_first_valid_index(X)
    return X.iloc[first_valid_index:]


def get_first_valid_index(series: Union[pd.Series, pd.DataFrame]) -> int:
    if series.empty:
        return 0
    if isinstance(series, pd.DataFrame):
        return next(
            (
                idx
                for idx, (_, x) in enumerate(series.iterrows())
                if not pd.isna(x).any()
            ),
            None,
        )
    elif isinstance(series, pd.Series):
        return next(
            (idx for idx, (_, x) in enumerate(series.items()) if not pd.isna(x)),
            None,
        )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

from __future__ import annotations

from inspect import getfullargspec
from typing import Callable, Optional, Type, Union

import pandas as pd
from sklearn.base import BaseEstimator

from ..base import Artifact, Tunable
from .base import Model


class WrapSKLearnClassifier(Model, Tunable):
    """
    Wraps an SKLearn Classifier model.
    There's no need to use it directly, `fold` automatically wraps all sklearn classifiers into this class.
    """

    def __init__(
        self,
        model_class: Type,
        init_args: dict,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.model = model_class(**init_args)
        self.params_to_try = params_to_try
        self.name = name or self.model.__class__.__name__
        self.properties = Model.Properties(
            requires_X=True, model_type=Model.Properties.ModelType.classifier
        )

    @classmethod
    def from_model(
        cls,
        model,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> WrapSKLearnClassifier:
        return cls(
            model_class=model.__class__,
            init_args=model.get_params(),
            params_to_try=params_to_try,
            name=name,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)
        else:
            fit_with_parameters(self.model, X, y, sample_weights)

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        probabilities = pd.DataFrame(
            data=self.model.predict_proba(X),
            index=X.index,
            columns=[
                f"probabilities_{self.name}_{item}" for item in self.model.classes_
            ],
        )
        predictions = pd.Series(
            data=self.model.predict(X).squeeze(),
            index=X.index,
            name=f"predictions_{self.name}",
        )
        return pd.concat([predictions, probabilities], copy=False, axis="columns")

    predict_in_sample = predict

    def get_params(self) -> dict:
        return self.model.get_params()

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        return WrapSKLearnClassifier(
            model_class=self.model.__class__,
            init_args=parameters,
            name=self.name,
        )


class WrapSKLearnRegressor(Model, Tunable):
    """
    Wraps an SKLearn regressor model.
    There's no need to use it directly, `fold` automatically wraps all sklearn regressors into this class.
    """

    def __init__(
        self,
        model_class: Type,
        init_args: dict,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.model = model_class(**init_args)
        self.params_to_try = params_to_try
        self.name = name or self.model.__class__.__name__
        self.properties = Model.Properties(
            requires_X=True, model_type=Model.Properties.ModelType.regressor
        )

    @classmethod
    def from_model(
        cls,
        model,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> WrapSKLearnRegressor:
        return cls(
            model_class=model.__class__,
            init_args=model.get_params(),
            name=name,
            params_to_try=params_to_try,
        )

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> Optional[Artifact]:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)
        else:
            fit_with_parameters(self.model, X, y, sample_weights)

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> Optional[Artifact]:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        return pd.Series(
            data=self.model.predict(X).squeeze(),
            index=X.index,
            name=f"predictions_{self.name}",
        ).to_frame()

    predict_in_sample = predict

    def get_params(self) -> dict:
        return self.model.get_params()

    def clone_with_params(
        self, parameters: dict, clone_children: Optional[Callable] = None
    ) -> Tunable:
        return WrapSKLearnRegressor(
            model_class=self.model.__class__,
            init_args=parameters,
            name=self.name,
        )


def fit_with_parameters(
    instance: BaseEstimator,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series],
) -> None:
    argspec = getfullargspec(instance.fit)  # type: ignore
    if len(argspec.args) == 1:
        instance.fit(X, y)  # type: ignore
    elif len(argspec.args) == 2:
        instance.fit(X, y, sample_weights)  # type: ignore
    else:
        raise ValueError(
            f"Expected 2 or 3 arguments for fit, but got {len(argspec.args)}."
        )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Callable, List, Optional

import pandas as pd
from sklearn.model_selection import ParameterGrid

from fold.base.scoring import score_results

from ..base import (
    Artifact,
    Optimizer,
    Pipeline,
    Tunable,
    get_concatenated_names,
    traverse_apply,
)
from ..splitters import SingleWindowSplitter
from ..utils.list import to_hierachical_dict, wrap_in_list
from .utils import _apply_params, _check_for_duplicate_names, _extract_param_grid

_divider = "¦"


class OptimizeGridSearch(Optimizer):
    candidates: Optional[List[Pipeline]] = None
    selected_params_: Optional[dict] = None
    selected_pipeline_: Optional[Tunable] = None
    scores_: Optional[List[float]] = None
    param_permutations: List[dict]

    def __init__(
        self,
        pipeline: Pipeline,
        krisi_metric_key: str,
        krisi_args: Optional[dict] = None,
        is_scorer_loss: bool = True,
        splitter: SingleWindowSplitter = SingleWindowSplitter(0.7),
        name: Optional[str] = None,
    ) -> None:
        self.pipeline = wrap_in_list(pipeline)
        self.name = name or "OptimizeGridSearch-" + get_concatenated_names(
            self.pipeline
        )
        self.krisi_metric_key = krisi_metric_key
        self.krisi_args = krisi_args or {}
        self.is_scorer_loss = is_scorer_loss
        self.splitter = splitter
        _check_for_duplicate_names(self.pipeline)

    @classmethod
    def from_cloned_instance(
        cls,
        pipeline: Pipeline,
        krisi_metric_key: str,
        krisi_args: Optional[dict],
        is_scorer_loss: bool,
        splitter: SingleWindowSplitter,
        candidates: Optional[List[Tunable]],
        selected_params_: Optional[dict],
        selected_pipeline_: Optional[Tunable],
        scores_: Optional[List[float]],
        name: Optional[str],
        id: str,
    ) -> OptimizeGridSearch:
        instance = cls(
            pipeline=pipeline,
            krisi_metric_key=krisi_metric_key,
            krisi_args=krisi_args,
            is_scorer_loss=is_scorer_loss,
            splitter=splitter,
        )
        instance.candidates = candidates
        instance.selected_params_ = selected_params_
        instance.selected_pipeline_ = selected_pipeline_
        instance.scores_ = scores_
        instance.name = name
        instance.id = id
        return instance

    def get_candidates(self) -> List[Pipeline]:
        if self.candidates is None:
            param_grid = _extract_param_grid(self.pipeline, _divider)

            self.param_permutations = [
                to_hierachical_dict(params, _divider)
                for params in ParameterGrid(param_grid)
            ]

            self.candidates = [
                traverse_apply(self.pipeline, _apply_params(params))
                for params in self.param_permutations
            ]
            return self.candidates
        else:
            return []

    def get_optimized_pipeline(self) -> Optional[Tunable]:
        return self.selected_pipeline_

    def process_candidate_results(
        self,
        results: List[pd.DataFrame],
        y: pd.Series,
        artifacts: List[pd.DataFrame],
    ) -> Optional[Artifact]:
        scores = [
            score_results(
                result,
                y=y[-len(result) :],
                artifacts=artifact,
                krisi_args=self.krisi_args,
            )[0][self.krisi_metric_key].result
            for result, artifact in zip(results, artifacts)
        ]
        selected_index = (
            scores.index(min(scores))
            if self.is_scorer_loss
            else scores.index(max(scores))
        )
        self.selected_params_ = self.param_permutations[selected_index]
        self.selected_pipeline_ = self.candidates[selected_index]
        # TODO: need to store the params for each score as well
        self.scores_ = scores
        return pd.DataFrame(
            {"selected_params": [self.selected_params_]}, index=y.index[-1:]
        )

    def clone(self, clone_children: Callable) -> OptimizeGridSearch:
        return OptimizeGridSearch.from_cloned_instance(
            pipeline=clone_children(self.pipeline),
            krisi_metric_key=self.krisi_metric_key,
            krisi_args=self.krisi_args,
            is_scorer_loss=self.is_scorer_loss,
            splitter=self.splitter,
            candidates=self.candidates,
            selected_params_=self.selected_params_,
            selected_pipeline_=self.selected_pipeline_,
            scores_=self.scores_,
            name=self.name,
            id=self.id,
        )



from collections import Counter
from copy import deepcopy
from typing import Callable, List, Union

from deepmerge import always_merger

from ..base import Clonable, Composite, Pipeline, Transformation, Tunable, traverse
from ..transformations.dev import Identity


def _process_params(params_to_try: dict) -> dict:
    if "passthrough" in params_to_try:
        params_to_try["passthrough"] = [True, False]
    return params_to_try


def _get_tunables_with_params_to_try(pipeline: Pipeline) -> List[Tunable]:
    return [
        i
        for i in traverse(pipeline)
        if isinstance(i, Tunable) and i.get_params_to_try() is not None
    ]


def _apply_params(params: dict) -> Callable:
    def __apply_params_to_transformation(
        item: Union[Composite, Transformation], clone_children: Callable
    ) -> Union[Composite, Transformation]:
        if not isinstance(item, Tunable):
            if isinstance(item, Clonable):
                return item.clone(clone_children)
            else:
                return item
        selected_params = params.get(item.name, {})
        if "passthrough" in selected_params and selected_params["passthrough"] is True:
            return Identity()  # type: ignore
        return item.clone_with_params(
            parameters=always_merger.merge(
                item.get_params(), _clean_params(selected_params)
            ),
            clone_children=clone_children,
        )  # type: ignore

    return __apply_params_to_transformation


def _clean_params(
    params_to_try: dict, keys: List[str] = ["passthrough", "_conditional"]
) -> dict:
    return {
        k: _clean_params(v) if isinstance(v, dict) else deepcopy(v)
        for k, v in params_to_try.items()
        if k not in keys
    }


def _check_for_duplicate_names(pipeline: Pipeline):
    names = [i.name for i in _get_tunables_with_params_to_try(pipeline)]
    if len(set(names)) != len(names):
        duplicate_names = [item for item, count in Counter(names).items() if count > 1]
        raise ValueError(
            f"Duplicate names in pipeline are not allowed. {duplicate_names}"
        )


def _extract_param_grid(pipeline: Pipeline, divider: str = "¦"):
    tunables = _get_tunables_with_params_to_try(pipeline)

    param_grid = {
        f"{transformation.name}{divider}{key}": value
        for transformation in tunables
        for key, value in _process_params(transformation.get_params_to_try()).items()
    }
    return param_grid



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import List, Optional, Tuple, Union

import pandas as pd

from ..base import Transformation, Tunable, feature_name_separator, fit_noop
from ..utils.checks import get_column_names, is_X_available
from ..utils.dataframe import to_dataframe
from ..utils.list import flatten, transform_range_to_list, wrap_in_list


class AddLagsY(Transformation, Tunable):
    """
    Adds past values of `y`.

    Parameters
    ----------
    lags : List[int], range
        A list of lags (of the target) to add as features.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddLagsY
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> _, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddLagsY([1,2,3])
    >>> preds, trained_pipeline = train_backtest(pipeline, None, y, splitter)
    >>> preds.head()
                         y_lag_1  y_lag_2  y_lag_3
    2021-12-31 15:40:00  -0.0000  -0.0126  -0.0251
    2021-12-31 15:41:00   0.0126  -0.0000  -0.0126
    2021-12-31 15:42:00   0.0251   0.0126  -0.0000
    2021-12-31 15:43:00   0.0377   0.0251   0.0126
    2021-12-31 15:44:00   0.0502   0.0377   0.0251

    ```
    """

    def __init__(
        self,
        lags: Union[List[int], range],
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        if not isinstance(lags, range) and not isinstance(lags, List):
            raise ValueError("lags must be a range or a List")
        self.lags = sorted(transform_range_to_list(lags))
        self.properties = Transformation.Properties(
            requires_X=False,
            mode=Transformation.Properties.Mode.online,
            memory_size=max(self.lags),
            _internal_supports_minibatch_backtesting=True,
        )
        self.params_to_try = params_to_try
        self.name = name or f"AddLagsY-{self.lags}"

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        lags = pd.DataFrame([], index=X.index)

        past_y = (
            self._state.memory_y if in_sample else self._state.memory_y.reindex(X.index)
        )
        lags = pd.concat(
            [past_y.shift(lag)[-len(X) :].rename(f"y_lag_{lag}") for lag in self.lags],
            copy=False,
            axis="columns",
        ).fillna(0.0)
        if is_X_available(X):
            return pd.concat([X, lags], copy=False, axis="columns")
        else:
            # If X is just an DataFrame with zeros, then just return the lags
            return lags

    fit = fit_noop
    update = fit_noop


class AddLagsX(Transformation, Tunable):
    """
    Adds past values of `X` for the desired column(s).

    Parameters
    ----------
    columns_and_lags : List[ColumnAndLag], ColumnAndLag
        A tuple (or a list of tuples) of the column name and a single or a list of lags to add as features.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddLagsX
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = AddLagsX([("sine", 1), ("sine", [2,3])])
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  sine~lag_1  sine~lag_2  sine~lag_3
    2021-12-31 15:40:00 -0.0000     -0.0126     -0.0251     -0.0377
    2021-12-31 15:41:00  0.0126     -0.0000     -0.0126     -0.0251
    2021-12-31 15:42:00  0.0251      0.0126     -0.0000     -0.0126
    2021-12-31 15:43:00  0.0377      0.0251      0.0126     -0.0000
    2021-12-31 15:44:00  0.0502      0.0377      0.0251      0.0126

    ```
    """

    ColumnAndLag = Tuple[str, Union[int, List[int]]]

    def __init__(
        self,
        columns_and_lags: Union[List[ColumnAndLag], ColumnAndLag],
        keep_original: bool = True,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.columns_and_lags = wrap_in_list(columns_and_lags)
        self.params_to_try = params_to_try

        def check_and_transform_if_needed(
            column_and_lag: AddLagsX.ColumnAndLag,
        ) -> AddLagsX.ColumnAndLag:
            column, lags = column_and_lag
            if (
                not isinstance(lags, int)
                and not isinstance(lags, List)
                and not isinstance(lags, range)
            ):
                raise ValueError("lags must be an int or a List or a range")
            lags = sorted(
                transform_range_to_list([lags] if isinstance(lags, int) else lags)
            )
            return column, lags

        self.columns_and_lags = list(
            map(check_and_transform_if_needed, self.columns_and_lags)
        )
        self.properties = Transformation.Properties(
            requires_X=True,
            memory_size=max(flatten([l for _, l in self.columns_and_lags])),
            disable_memory=True,
        )
        self.keep_original = keep_original
        self.name = name or f"AddLagsX-{self.columns_and_lags}"

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        lagged_columns = []
        for column, lags in self.columns_and_lags:
            for lag in lags:
                selected = to_dataframe(X[get_column_names(column, X)])
                lagged_columns.append(
                    selected.shift(lag)[-len(X) :].add_suffix(
                        f"{feature_name_separator}lag_{lag}"
                    )
                )
        to_concat = [X] + lagged_columns if self.keep_original else lagged_columns
        return pd.concat(to_concat, copy=False, axis="columns").fillna(0.0)

    fit = fit_noop
    update = fit_noop



# from __future__ import annotations

# from copy import deepcopy
# from typing import Any, Dict, Optional, Type, Union

# import numpy as np
# import pandas as pd

# from fold.models.base import Model


# class WrapNeuralForecast(Model):
#     def __init__(
#         self,
#         model_class: Type,
#         init_args: Optional[Dict],
#         instance: Optional[Any] = None,
#     ) -> None:
#         self.init_args = init_args
#         init_args = {} if init_args is None else init_args
#         self.model = model_class(**init_args) if instance is None else instance
#         self.model_class = model_class
#         self.name = self.model_class.__class__.__name__
#         self.properties = Model.Properties(
#             requires_X=False,
#             model_type=Model.Properties.ModelType.regressor,
#         )

# @classmethod
# def from_model(
#     cls,
#     model,
# ) -> WrapNeuralForecast:
#     return cls(
#         model_class=None,
#         init_args=None,
#         instance=model,
#     )

# def fit(
#     self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
# ) -> None:
#     from neuralforecast import NeuralForecast

#     self.nf = NeuralForecast(models=[self.model], freq=X.index.freqstr)
#     assert type(self.model.h) is int, "Forecasting horizon/step must be an integer."

#     data = y.rename("y").to_frame()
#     data["ds"] = X.index
#     data["unique_id"] = 1.0
#     self.nf.fit(data)

# def update(
#     self,
#     X: pd.DataFrame,
#     y: Optional[pd.Series],
#     sample_weights: Optional[pd.Series] = None,
# ) -> None:
#     for model in self.nf.models:
#         model.max_steps = 10
#     data = y.rename("y").to_frame()
#     data["ds"] = X.index
#     data["unique_id"] = 1.0
#     self.nf.fit(data)

# def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
#     predicted = self.nf.predict()

#     if len(predicted) != len(X):
#         raise ValueError(
#             "Step size (of the Splitter) and `h` (forecasting horizon) must be"
#             " equal."
#         )
#     else:
#         return pd.Series(
#             predicted[self.model.__class__.__name__].values, index=X.index
#         )

# def predict_in_sample(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
#     data = pd.DataFrame(
#         {"ds": X.index, "y": 0.0, "unique_id": 1.0},
#     )
#     nf = deepcopy(self.nf)
#     predictions = nf.predict_rolled(
#         data,
#         insample_size=len(X) - self.model.input_size,
#         n_windows=None,
#         step_size=self.model.input_size,
#     )[self.model.__class__.__name__]
#     # NeuralForecast will not return in sample predictions for `input_size`, so let's pad that with NaNs
#     padding_size = len(X) - len(predictions)
#     return pd.Series(
#         np.hstack([np.full(padding_size, np.nan), predictions.values]),
#         index=X.index,
#     )



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Callable, List, Optional

import pandas as pd
from tqdm import tqdm
from tqdm.auto import tqdm as tqdm_auto

from ...base import Artifact, Composite, Pipeline, TrainedPipeline, X
from ...splitters import Fold
from ..types import Backend, Stage

DEBUG_MULTI_PROCESSING = False


def train_pipeline(
    self,
    func: Callable,
    pipeline: Pipeline,
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    splits: List[Fold],
    never_update: bool,
    backend: Backend,
    project_name: str,
    silent: bool,
):
    if DEBUG_MULTI_PROCESSING:
        from ray.cloudpickle import dumps, loads

        pipeline = loads(dumps(pipeline))
    return [
        func(X, y, artifact, pipeline, split, never_update, backend, project_name)
        for split in tqdm_auto(splits, desc="Training", disable=silent)
    ]


def backtest_pipeline(
    self,
    func: Callable,
    pipeline: TrainedPipeline,
    splits: List[Fold],
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    backend: Backend,
    mutate: bool,
    silent: bool,
):
    if DEBUG_MULTI_PROCESSING:
        from ray.cloudpickle import dumps, loads

        pipeline = loads(dumps(pipeline))
    return [
        func(pipeline, split, X, y, artifact, backend, mutate)
        for split in tqdm(splits, desc="Backtesting", disable=silent)
    ]


def process_child_transformations(
    self,
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: X,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
):
    if DEBUG_MULTI_PROCESSING:
        from ray.cloudpickle import dumps, loads

        list_of_child_transformations_with_index = loads(
            dumps(list_of_child_transformations_with_index)
        )
    return [
        func(
            composite,
            index,
            child_transformation,
            X,
            y,
            artifacts,
            stage,
            backend,
            results_primary,
            tqdm,
        )
        for index, child_transformation in list_of_child_transformations_with_index
    ]


class NoBackend(Backend):
    name = "no"
    process_child_transformations = process_child_transformations
    train_pipeline = train_pipeline
    backtest_pipeline = backtest_pipeline



from functools import reduce
from typing import Callable, List, Optional, Union

import numpy as np
import pandas as pd

from .enums import ParsableEnum
from .list import filter_none, flatten, has_intersection, keep_only_duplicates


def to_series(dataframe_or_series: Union[pd.DataFrame, pd.Series]) -> pd.Series:
    if isinstance(dataframe_or_series, pd.DataFrame):
        if len(dataframe_or_series.columns) != 1:
            raise ValueError("DataFrame must have exactly one column")
        if len(dataframe_or_series) == 1:
            return dataframe_or_series[dataframe_or_series.columns[0]]
        else:
            return dataframe_or_series.squeeze()
    elif isinstance(dataframe_or_series, pd.Series):
        return dataframe_or_series
    else:
        raise ValueError("Not a pd.Series or pd.DataFrame")


def to_dataframe(dataframe_or_series: Union[pd.DataFrame, pd.Series]) -> pd.DataFrame:
    if isinstance(dataframe_or_series, pd.DataFrame):
        return dataframe_or_series
    elif isinstance(dataframe_or_series, pd.Series):
        return dataframe_or_series.to_frame()
    else:
        raise ValueError("Not a pd.Series or pd.DataFrame")


def __concat_on_axis(axis: str) -> Callable:
    def concat_on(
        dfs: List[Optional[Union[pd.DataFrame, pd.Series]]], copy: bool
    ) -> pd.DataFrame:
        filtered = filter_none(dfs)
        if len(filtered) == 0:
            return None  # type: ignore
        elif len(filtered) == 1:
            return filtered[0]
        else:
            return pd.concat(filtered, axis=axis, copy=copy)

    return concat_on


concat_on_columns = __concat_on_axis("columns")
concat_on_index = __concat_on_axis("index")


def concat_on_index_override_duplicate_rows(dfs: List[pd.DataFrame]) -> pd.DataFrame:
    if len(dfs) == 0:
        return pd.DataFrame()
    elif len(dfs) == 1:
        return dfs[0]
    else:
        return pd.concat(dfs, axis="index", copy=False).groupby(level=0).last()


def take_log(df: pd.DataFrame) -> pd.DataFrame:
    """
    If you use np.log() on a DataFrame with a single column, it'll turn it into a pd.Series.
    This function retains the original instance type (pd.DataFrame).
    """
    result = np.log(df)
    if isinstance(df, pd.Series):
        return result.to_frame()
    else:
        return result


class ResolutionStrategy(ParsableEnum):
    """
    Parameters
    ----------
    first : str
        Only keep the first (leftmost) duplicate column(s).
    last : str
        Only keep the last (rightmost) duplicate column(s).
    both : str
        Keep both duplicate columns.
    """

    first = "first"
    last = "last"
    both = "both"


def concat_on_columns_with_duplicates(
    dfs: List[Union[pd.DataFrame, pd.Series]],
    strategy: ResolutionStrategy,
    raise_indices_dont_match: bool = False,
) -> pd.DataFrame:
    if len(dfs) == 0:
        return pd.DataFrame()
    if all([result.empty for result in dfs]):
        return pd.DataFrame(index=dfs[0].index)
    dfs = [to_dataframe(df) for df in dfs]
    max_len = max([len(df) for df in dfs])
    if not all([len(df) == max_len for df in dfs]):
        index_to_apply = reduce(
            lambda i1, i2: i1.union(i2), [df.index for df in dfs]
        ).sort_values()
        dfs = [df.reindex(index_to_apply) for df in dfs]

    columns = flatten([result.columns.to_list() for result in dfs])
    duplicates = keep_only_duplicates(columns)
    if len(duplicates) == 0:
        return pd.concat(dfs, axis="columns", copy=False)

    if len(duplicates) > 0 or strategy is not ResolutionStrategy.both:

        def concat(
            dfs: List[pd.DataFrame],
        ) -> pd.DataFrame:
            if not raise_indices_dont_match and not all(
                [df.index.equals(dfs[0].index) for df in dfs]
            ):
                return reduce(lambda accum, item: accum.combine_first(item), dfs)
            else:
                return pd.concat(dfs, copy=False, axis="columns")

        duplicate_columns = [
            result[
                keep_only_duplicates(flatten([duplicates, result.columns.to_list()]))
            ]
            for result in dfs
            if has_intersection(result.columns.to_list(), duplicates)
        ]
        results = [result.drop(columns=duplicates, errors="ignore") for result in dfs]
        results = [
            df for df in results if not df.empty
        ]  # if all of them are empty, create an empty list

        if strategy is ResolutionStrategy.first:
            return concat(results + [duplicate_columns[0]])
        elif strategy is ResolutionStrategy.last:
            return concat(
                results + [duplicate_columns[-1]],
            )
        elif strategy is ResolutionStrategy.both:
            return concat(results + duplicate_columns)
        else:
            raise ValueError(f"ResolutionStrategy is not valid: {strategy}")
    else:
        return pd.concat(dfs, copy=False, axis="columns")


def fill_na_inf(df: pd.DataFrame) -> pd.DataFrame:
    return df.replace([np.inf, -np.inf], 0.0).fillna(0.0)


def apply_function_batched(
    df: pd.DataFrame, func: Callable, batch_columns: Optional[int]
) -> pd.DataFrame:
    if batch_columns is not None:
        batches = [
            func(df.iloc[:, i : i + batch_columns])
            for i in range(0, len(df.columns), batch_columns)
        ]
        return pd.concat(batches, axis="columns", copy=False)
    else:
        return func(df)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Callable, Optional

import pandas as pd

from ..base import Transformation, Tunable, fit_noop
from ..utils.dataframe import apply_function_batched, fill_na_inf


class ApplyFunction(Transformation, Tunable):
    """
    Wraps and arbitrary function that will run at inference.
    """

    def __init__(
        self,
        func: Callable,
        past_window_size: Optional[int],
        fillna: bool = False,
        batch_columns: Optional[int] = None,
        output_dtype: Optional[type] = None,
        name: Optional[str] = None,
        params_to_try: Optional[dict] = None,
    ) -> None:
        self.func = func
        self.name = name or func.__name__
        self.fillna = fillna
        self.past_window_size = past_window_size
        self.batch_columns = batch_columns
        self.output_dtype = output_dtype
        self.params_to_try = params_to_try
        self.properties = Transformation.Properties(
            requires_X=True, memory_size=past_window_size, disable_memory=True
        )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        def convert_dtype_if_needed(df: pd.DataFrame) -> pd.DataFrame:
            return df.astype(self.output_dtype) if self.output_dtype is not None else df

        return_value = convert_dtype_if_needed(
            apply_function_batched(X, self.func, self.batch_columns)
        )
        return fill_na_inf(return_value) if self.fillna else return_value

    fit = fit_noop
    update = fit



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from copy import deepcopy
from typing import Callable, List, Tuple, TypeVar

import pandas as pd

from ..base import Block, Clonable, Composite, Pipeline, TrainedPipelines
from ..base.utils import get_maximum_memory_size, traverse, traverse_apply
from ..splitters import Fold
from .types import Stage


def deepcopy_pipelines(transformation: Pipeline) -> Pipeline:
    if isinstance(transformation, List) or isinstance(transformation, Tuple):
        return [deepcopy_pipelines(t) for t in transformation]
    elif isinstance(transformation, Clonable):
        return transformation.clone(deepcopy_pipelines)
    else:
        return deepcopy(transformation)


def replace_with(transformations: List[Block]) -> Callable:
    mapping = {t.id: t for t in traverse(transformations)}

    def replace(transformation: Pipeline) -> Pipeline:
        if isinstance(transformation, List) or isinstance(transformation, Tuple):
            return [replace(t) for t in transformation]

        if transformation.id in mapping:
            return mapping[transformation.id]

        if isinstance(transformation, Clonable):
            return transformation.clone(replace)

        return transformation

    return replace


def _extract_trained_pipelines(
    processed_idx: List[int], processed_pipelines: List[Pipeline]
) -> TrainedPipelines:
    return [
        pd.Series(
            transformation_over_time,
            index=processed_idx,
            name=transformation_over_time[0].name,
        )
        for transformation_over_time in zip(*processed_pipelines)
    ]


def _set_metadata(
    pipeline: Pipeline,
    metadata: Composite.Metadata,
) -> Pipeline:
    def set_(block: Block, clone_children: Callable) -> Block:
        if isinstance(block, Clonable):
            block = block.clone(clone_children)
            if isinstance(block, Composite):
                block.metadata = metadata
        return block

    return traverse_apply(pipeline, set_)


T = TypeVar("T")


def _cut_to_train_window(df: T, fold: Fold, stage: Stage) -> T:
    window_start = (
        fold.update_window_start if stage == Stage.update else fold.train_window_start
    )
    window_end = (
        fold.update_window_end if stage == Stage.update else fold.train_window_end
    )
    return df.iloc[window_start:window_end]  # type: ignore


def _cut_to_backtesting_window(df: T, fold: Fold, pipeline: Pipeline) -> T:
    overlap = get_maximum_memory_size(pipeline)
    test_window_start = max(fold.test_window_start - overlap, 0)
    return df.iloc[test_window_start : fold.test_window_end]  # type: ignore



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.

from .classes import *  # noqa
from .utils import *

feature_name_separator = "~"



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from .loop.backtesting import backtest
from .loop.encase import backtest_score, train_backtest, train_evaluate
from .loop.training import train
from .loop.types import Backend, BackendType, TrainMethod
from .splitters import (
    ExpandingWindowSplitter,
    SingleWindowSplitter,
    SlidingWindowSplitter,
)



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Callable

from ..utils.enums import ParsableEnum


@dataclass
class Backend:
    name: str
    process_child_transformations: Callable
    train_pipeline: Callable
    backtest_pipeline: Callable

    def __init__(self):  # to be able to initalize it with the default parameters
        pass


class BackendType(ParsableEnum):
    """
    Parameters
    ----------
    no: string
        Uses sequential processing. This is the default.
    ray: string
        Uses `ray` as a backend. Call `ray.init()` before using this backend.
    pathos: string
        Uses `pathos.multiprocessing` as a backend (via [p_tqdm](https://github.com/swansonk14/p_tqdm)).
    thread: string
        Uses `threading` as a backend (via tqdm.contrib.concurrent.thread_map).

    """

    no = "no"
    ray = "ray"
    pathos = "pathos"
    thread = "thread"
    joblib = "joblib"


class TrainMethod(ParsableEnum):
    """
    Parameters
    ----------
    parallel: string
        Parallel, independent training of pipelines for each fold.
    sequential: string
        Sequentially train/update pipelines, walking forward in time.

    """

    parallel = "parallel"
    sequential = "sequential"
    parallel_with_search = (  # Don't use it just yet, not yet fully documented
        "parallel_with_search"
    )


class Stage(Enum):
    inital_fit = "inital_fit"
    update = "update"
    update_online_only = "update_online_only"
    infer = "infer"

    def is_fit_or_update(self) -> bool:
        return self in [Stage.inital_fit, Stage.update]



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from __future__ import annotations

from typing import Optional, Union

import pandas as pd

from ..base import Tunable, fit_noop
from .base import TimeSeriesModel


class Naive(TimeSeriesModel):
    """
    A univariate model that predicts the last target value.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.models import Naive
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(train_window=0.5, step=0.2)
    >>> pipeline = Naive()
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> pd.concat([preds, y[preds.index]], axis=1).head()
                         predictions_Naive    sine
    2021-12-31 15:40:00            -0.0000  0.0126
    2021-12-31 15:41:00             0.0126  0.0251
    2021-12-31 15:42:00             0.0251  0.0377
    2021-12-31 15:43:00             0.0377  0.0502
    2021-12-31 15:44:00             0.0502  0.0628

    ```
    """

    name = "Naive"
    properties = TimeSeriesModel.Properties(
        requires_X=False,
        mode=TimeSeriesModel.Properties.Mode.online,
        memory_size=1,
        _internal_supports_minibatch_backtesting=True,
    )

    def predict(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        # it's an online transformation, so len(X) will be always 1,
        return pd.Series(past_y.iloc[-1].squeeze(), index=X.index[-1:None]).fillna(0.0)

    def predict_in_sample(
        self, X: pd.DataFrame, lagged_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        return lagged_y.fillna(0.0)

    fit = fit_noop
    update = fit


class NaiveSeasonal(TimeSeriesModel, Tunable):
    """
    A model that predicts the last value seen in the same season.
    """

    name = "NaiveSeasonal"

    def __init__(
        self, seasonal_length: int, params_to_try: Optional[dict] = None
    ) -> None:
        assert seasonal_length > 1, "seasonal_length must be greater than 1"
        self.seasonal_length = seasonal_length
        self.properties = TimeSeriesModel.Properties(
            requires_X=False,
            mode=TimeSeriesModel.Properties.Mode.online,
            memory_size=seasonal_length,
            _internal_supports_minibatch_backtesting=True,
        )
        self.params_to_try = params_to_try

    def predict(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        # it's an online transformation, so len(X) will be always 1,
        return pd.Series(
            past_y.iloc[-self.seasonal_length].squeeze(),
            index=X.index[-1:None],
        ).fillna(0.0)

    def predict_in_sample(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        return past_y.shift(self.seasonal_length - 1).fillna(0.0)

    fit = fit_noop
    update = fit

    def get_params(self) -> dict:
        return {"seasonal_length": self.seasonal_length}


class MovingAverage(TimeSeriesModel, Tunable):
    """
    A model that predicts the mean of the last values seen.
    """

    name = "MovingAverage"

    def __init__(self, window_size: int, params_to_try: Optional[dict] = None) -> None:
        self.window_size = window_size
        self.properties = TimeSeriesModel.Properties(
            requires_X=False,
            mode=TimeSeriesModel.Properties.Mode.online,
            memory_size=window_size,
            _internal_supports_minibatch_backtesting=True,
        )
        self.params_to_try = params_to_try

    def predict(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        # it's an online transformation, so len(X) will be always 1,
        return pd.Series(
            past_y.rolling(self.window_size, min_periods=0).mean()[-1],
            index=X.index[-1:None],
        ).fillna(0.0)

    def predict_in_sample(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        return past_y.rolling(self.window_size, min_periods=0).mean().fillna(0.0)

    fit = fit_noop
    update = fit

    def get_params(self) -> dict:
        return {"window_size": self.window_size}


class ExponentiallyWeightedMovingAverage(TimeSeriesModel, Tunable):
    """
    A model that predicts the exponentially weighed mean of the last values seen.
    """

    name = "ExponentiallyWeightedMovingAverage"

    def __init__(self, window_size: int, params_to_try: Optional[dict] = None) -> None:
        self.window_size = window_size
        self.properties = TimeSeriesModel.Properties(
            requires_X=False,
            mode=TimeSeriesModel.Properties.Mode.online,
            memory_size=self.window_size * 4,
            _internal_supports_minibatch_backtesting=True,
        )
        self.params_to_try = params_to_try

    def predict(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        # it's an online transformation, so len(X) will be always 1,
        return pd.Series(
            past_y.ewm(alpha=1 / self.window_size, adjust=True, min_periods=0).mean()[
                -1
            ],
            index=X.index[-1:None],
        ).fillna(0.0)

    def predict_in_sample(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        return (
            past_y.ewm(alpha=1 / self.window_size, adjust=True, min_periods=0)
            .mean()
            .fillna(0.0)
        )

    fit = fit_noop
    update = fit

    def get_params(self) -> dict:
        return {"window_size": self.window_size}



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import Callable, List, Optional

import pandas as pd
from tqdm import tqdm
from tqdm.contrib.concurrent import thread_map

from ...base import Artifact, Composite, Pipeline, TrainedPipeline, X
from ...splitters import Fold
from ..types import Backend, Stage


def train_pipeline(
    self,
    func: Callable,
    pipeline: Pipeline,
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    splits: List[Fold],
    never_update: bool,
    backend: Backend,
    project_name: str,
    silent: bool,
):
    return thread_map(
        lambda split: func(
            X, y, artifact, pipeline, split, never_update, backend, project_name
        ),
        splits,
        disable=silent,
    )


def backtest_pipeline(
    self,
    func: Callable,
    pipeline: TrainedPipeline,
    splits: List[Fold],
    X: pd.DataFrame,
    y: pd.Series,
    artifact: Artifact,
    backend: Backend,
    mutate: bool,
    silent: bool,
):
    return thread_map(
        lambda split: func(pipeline, split, X, y, artifact, backend, mutate),
        splits,
        disable=silent,
    )


def process_child_transformations(
    self,
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: X,
    y: Optional[pd.Series],
    artifacts: Artifact,
    stage: Stage,
    backend: Backend,
    results_primary: Optional[List[pd.DataFrame]],
    tqdm: Optional[tqdm] = None,
):
    list_of_child_transformations_with_index = [
        {"index": index, "child_transformation": child_transformation}
        for index, child_transformation in list_of_child_transformations_with_index
    ]
    return thread_map(
        lambda obj: func(
            composite,
            obj["index"],
            obj["child_transformation"],
            X,
            y,
            artifacts,
            stage,
            backend,
            results_primary,
            tqdm,
        ),
        list_of_child_transformations_with_index,
        disable=True,
    )


class ThreadBackend(Backend):
    name = "thread"
    process_child_transformations = process_child_transformations
    train_pipeline = train_pipeline
    backtest_pipeline = backtest_pipeline



# Copyright (c) 2022 - Present Myalo UG (haftungbeschränkt) (Mark Aron Szulyovszky, Daniel Szemerey) <info@dreamfaster.ai>. All rights reserved. See LICENSE in root folder.


from typing import List, Union

import pandas as pd

from fold.utils.dataframe import to_series
from fold.utils.list import flatten, unique


def is_prediction(input: pd.DataFrame) -> bool:
    if len(input.columns) == 1:
        return (
            input.columns[0].startswith("predictions_")
            if isinstance(input.columns[0], str)
            else False
        )
    else:
        is_predictions_col_present = input.columns[0].startswith("predictions_")
        return is_predictions_col_present and all(
            [col.startswith("probabilities_") for col in input.columns[1:]]
        )


def all_have_probabilities(results: List[pd.DataFrame]) -> bool:
    """
    Check if all the DataFrames have probabilities columns,
    or if their values are all NaN, indicating an empty DataFrame
    being passed around, as SkipNA filtered out all data.
    """
    return all(
        [
            any([True for col in df.columns if col.startswith("probabilities_")])
            or len(df.columns) == 0
            or (len(df.columns) == 1 and df.isna().sum()[0] == len(df))
            for df in results
        ]
    )


def has_probabilities(df: pd.DataFrame) -> bool:
    """
    Check if all the DataFrames have probabilities columns,
    or if their values are all NaN, indicating an empty DataFrame
    being passed around, as SkipNA filtered out all data.
    """
    return (
        any([True for col in df.columns if str(col).startswith("probabilities_")])
        or len(df.columns) == 0
        or (len(df.columns) == 1 and df.isna().sum()[0] == len(df))
    )


def get_prediction_column(input: pd.DataFrame) -> pd.Series:
    return to_series(input[get_prediction_column_name(input)])


def get_probabilities_columns(input: pd.DataFrame) -> pd.DataFrame:
    return input[get_probabilities_column_names(input)]


def get_probabilities_column_names(input: pd.DataFrame) -> List[str]:
    candidates = [col for col in input.columns if str(col).startswith("probabilities_")]
    if len(candidates) == 0:
        raise ValueError(f"Could not find any probabilities column in {input.columns}.")
    return candidates


def get_classes_from_probabilies_column_names(columns: List[str]) -> List[str]:
    return [col.split("_")[-1] for col in columns]


def get_prediction_column_name(input: pd.DataFrame) -> str:
    candidates = [col for col in input.columns if col.startswith("predictions_")]
    if len(candidates) == 0:
        if len(input.columns) == 1:
            return input.columns[0]
        else:
            raise ValueError(f"Could not find a predictions column in {input.columns}.")
    return candidates[0]


def is_X_available(X: pd.DataFrame) -> bool:
    """
    Check if X is available, or the input is univariate, without exogenous variables.
    """
    return not (X.shape[1] == 1 and X.columns[0] == "X_not_available")


def get_column_names(column_pattern: str, X: pd.DataFrame) -> pd.Index:
    if column_pattern == "all":
        return X.columns
    if column_pattern.endswith("*"):
        to_match = column_pattern.split("*")[0]
        return [col for col in X.columns if col.startswith(to_match)]
    if column_pattern.startswith("*"):
        to_match = column_pattern.split("*")[1]
        return [col for col in X.columns if col.endswith(to_match)]
    else:
        return [column_pattern]


def get_list_column_names(
    columns: List[str], X: pd.DataFrame
) -> Union[List[str], pd.Index]:
    assert isinstance(columns, list)
    if len(columns) == 1:
        return get_column_names(columns[0], X)
    else:
        return unique(flatten([get_column_names(c, X) for c in columns]))
